# Copyright (c) 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the GlusterFS driver module."""

import errno
import os
import six
import tempfile
import time
import traceback

import mock
import os_brick
from oslo_concurrency import processutils as putils
from oslo_config import cfg
from oslo_utils import imageutils
from oslo_utils import units

from cinder import compute
from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import test
from cinder import utils
from cinder.volume import driver as base_driver
from cinder.volume.drivers import glusterfs
from cinder.volume.drivers import remotefs as remotefs_drv


CONF = cfg.CONF


class DumbVolume(object):
    fields = {}

    def __setitem__(self, key, value):
        self.fields[key] = value

    def __getitem__(self, item):
        return self.fields[item]


class FakeDb(object):
    msg = "Tests are broken: mock this out."

    def volume_get(self, *a, **kw):
        raise Exception(self.msg)

    def snapshot_get_all_for_volume(self, *a, **kw):
        """Mock this if you want results from it."""
        return []


class GlusterFsDriverTestCase(test.TestCase):
    """Test case for GlusterFS driver."""

    TEST_EXPORT1 = 'glusterfs-host1:/export'
    TEST_EXPORT2 = 'glusterfs-host2:/export'
    TEST_EXPORT2_OPTIONS = '-o backupvolfile-server=glusterfs-backup1'
    TEST_SIZE_IN_GB = 1
    TEST_MNT_POINT = '/mnt/glusterfs'
    TEST_MNT_POINT_BASE = '/mnt/test'
    TEST_LOCAL_PATH = '/mnt/glusterfs/volume-123'
    TEST_FILE_NAME = 'test.txt'
    TEST_SHARES_CONFIG_FILE = '/etc/cinder/test-shares.conf'
    TEST_TMP_FILE = '/tmp/tempfile'
    VOLUME_UUID = 'abcdefab-cdef-abcd-efab-cdefabcdefab'
    VOLUME_NAME = 'volume-%s' % VOLUME_UUID
    SNAP_UUID = 'bacadaca-baca-daca-baca-dacadacadaca'
    SNAP_UUID_2 = 'bebedede-bebe-dede-bebe-dedebebedede'

    def setUp(self):
        super(GlusterFsDriverTestCase, self).setUp()
        self._configuration = mock.MagicMock()
        self._configuration.glusterfs_shares_config = \
            self.TEST_SHARES_CONFIG_FILE
        self._configuration.glusterfs_mount_point_base = \
            self.TEST_MNT_POINT_BASE
        self._configuration.nas_secure_file_permissions = 'false'
        self._configuration.nas_secure_file_operations = 'false'
        self._configuration.nas_ip = None
        self._configuration.nas_share_path = None
        self._configuration.nas_mount_options = None

        self._driver =\
            glusterfs.GlusterfsDriver(configuration=self._configuration,
                                      db=FakeDb())
        self._driver.shares = {}
        compute.API = mock.MagicMock()

    def assertRaisesAndMessageMatches(
            self, excClass, msg, callableObj, *args, **kwargs):
        """Ensure that 'excClass' was raised and its message contains 'msg'."""

        caught = False
        try:
            callableObj(*args, **kwargs)
        except Exception as exc:
            caught = True
            self.assertEqual(excClass, type(exc),
                             'Wrong exception caught: %s Stacktrace: %s' %
                             (exc, traceback.format_exc()))
            self.assertIn(msg, six.text_type(exc))

        if not caught:
            self.fail('Expected raised exception but nothing caught.')

    def test_local_path(self):
        """local_path common use case."""
        self.override_config("glusterfs_mount_point_base",
                             self.TEST_MNT_POINT_BASE)
        drv = self._driver

        volume = DumbVolume()
        volume['id'] = self.VOLUME_UUID
        volume['provider_location'] = self.TEST_EXPORT1
        volume['name'] = 'volume-123'

        self.assertEqual(
            '/mnt/test/ab03ab34eaca46a5fb81878f7e9b91fc/volume-123',
            drv.local_path(volume))

    def test_mount_glusterfs(self):
        """_mount_glusterfs common case usage."""
        drv = self._driver

        with mock.patch.object(os_brick.remotefs.remotefs.RemoteFsClient,
                               'mount') as mock_mount:
            drv._mount_glusterfs(self.TEST_EXPORT1)

            mock_mount.assert_called_once_with(self.TEST_EXPORT1, [])

    def test_mount_glusterfs_should_reraise_exception_on_failure(self):
        """_mount_glusterfs should reraise exception if mount fails."""
        drv = self._driver

        with mock.patch.object(os_brick.remotefs.remotefs.RemoteFsClient,
                               'mount') as mock_mount:

            mock_mount.side_effect = exception.GlusterfsException()

            self.assertRaises(exception.GlusterfsException,
                              drv._mount_glusterfs, self.TEST_EXPORT1)

    def test_get_hash_str(self):
        """_get_hash_str should calculation correct value."""
        drv = self._driver

        self.assertEqual('ab03ab34eaca46a5fb81878f7e9b91fc',
                         drv._get_hash_str(self.TEST_EXPORT1))

    def test_get_mount_point_for_share(self):
        """_get_mount_point_for_share should call RemoteFsClient."""
        drv = self._driver
        hashed_path = '/mnt/test/abcdefabcdef'

        with mock.patch.object(os_brick.remotefs.remotefs.RemoteFsClient,
                               'get_mount_point') as mock_get_mount_point:
            mock_get_mount_point.return_value = hashed_path

            result = drv._get_mount_point_for_share(self.TEST_EXPORT1)

            self.assertEqual(hashed_path, result)

    def test_get_available_capacity_with_df(self):
        """_get_available_capacity should calculate correct value."""
        drv = self._driver

        df_total_size = 2620544
        df_avail = 1490560
        df_head = 'Filesystem 1K-blocks Used Available Use% Mounted on\n'
        df_data = 'glusterfs-host:/export %d 996864 %d 41%% /mnt' % \
                  (df_total_size, df_avail)
        df_output = df_head + df_data

        with mock.patch.object(drv, '_get_mount_point_for_share') as \
                mock_get_mount_point_for_share,\
                mock.patch.object(drv, '_execute') as mock_execute:
            mock_get_mount_point_for_share.\
                return_value = self.TEST_MNT_POINT
            mock_execute.return_value = (df_output, None)

            result = drv._get_available_capacity(self.TEST_EXPORT1)
            self.assertEqual((df_avail, df_total_size), result)

    def test_get_provisioned_capacity(self):
        """_get_provisioned_size should calculate correct value."""
        drv = self._driver

        drv.shares = {'127.7.7.7:/gluster1': None}
        with mock.patch.object(drv, '_get_mount_point_for_share') as \
                mock_get_mount_point_for_share,\
                mock.patch.object(drv, '_execute') as mock_execute:
            mock_get_mount_point_for_share.\
                return_value = self.TEST_MNT_POINT
            mock_execute.return_value = ("3221225472 /mount/point", '')
            provisioned_capacity = drv._get_provisioned_capacity()

            self.assertEqual(3.0, provisioned_capacity)

    def test_update_volume_stats_thin(self):
        """_update_volume_stats_thin with qcow2 files."""
        drv = self._driver
        rfsdriver = remotefs_drv.RemoteFSSnapDriver

        with mock.patch.object(rfsdriver, '_update_volume_stats') as \
                mock_update_volume_stats,\
                mock.patch.object(drv, '_get_provisioned_capacity') as \
                mock_get_provisioned_capacity:
            data = {'total_capacity_gb': 10.0,
                    'free_capacity_gb': 2.0}
            drv._stats = data
            drv.configuration.nas_volume_prov_type = 'thin'
            drv.configuration.max_over_subscription_ratio = 20.0
            mock_get_provisioned_capacity.return_value = 8.0
            drv._update_volume_stats()
            data['max_over_subscription_ratio'] = 20.0
            data['thick_provisioning_support'] = False
            data['thin_provisioning_support'] = True

            self.assertEqual(data, drv._stats)
            self.assertTrue(mock_get_provisioned_capacity.called)
            self.assertTrue(mock_update_volume_stats.called)

    def test_update_volume_stats_thick(self):
        """_update_volume_stats_thick with raw files."""
        drv = self._driver
        rfsdriver = remotefs_drv.RemoteFSSnapDriver

        with mock.patch.object(rfsdriver, '_update_volume_stats') as \
                mock_update_volume_stats:
            data = {'total_capacity_gb': 10.0,
                    'free_capacity_gb': 2.0}
            drv._stats = data
            drv.configuration.nas_volume_prov_type = 'thick'
            drv.configuration.max_over_subscription_ratio = 20.0
            drv._update_volume_stats()
            data['provisioned_capacity_gb'] = 8.0
            data['max_over_subscription_ratio'] = 20.0
            data['thick_provisioning_support'] = True
            data['thin_provisioning_support'] = False

            self.assertEqual(data, drv._stats)
            self.assertTrue(mock_update_volume_stats.called)

    def test_load_shares_config(self):
        drv = self._driver

        drv.configuration.glusterfs_shares_config = (
            self.TEST_SHARES_CONFIG_FILE)

        with mock.patch.object(drv, '_read_config_file') as \
                mock_read_config_file:
            config_data = []
            config_data.append(self.TEST_EXPORT1)
            config_data.append('#' + self.TEST_EXPORT2)
            config_data.append(self.TEST_EXPORT2 + ' ' +
                               self.TEST_EXPORT2_OPTIONS)
            config_data.append('broken:share_format')
            config_data.append('')
            mock_read_config_file.return_value = config_data

            drv._load_shares_config(drv.configuration.glusterfs_shares_config)

            self.assertIn(self.TEST_EXPORT1, drv.shares)
            self.assertIn(self.TEST_EXPORT2, drv.shares)
            self.assertEqual(2, len(drv.shares))

            self.assertEqual(self.TEST_EXPORT2_OPTIONS,
                             drv.shares[self.TEST_EXPORT2])

    def test_ensure_share_mounted(self):
        """_ensure_share_mounted simple use case."""
        drv = self._driver
        with mock.patch.object(utils, 'get_file_mode') as \
                mock_get_file_mode,\
                mock.patch.object(utils, 'get_file_gid') as mock_get_file_gid,\
                mock.patch.object(drv, '_execute') as mock_execute,\
                mock.patch.object(drv, '_ensure_share_writable') as \
                mock_ensure_share_writable,\
                mock.patch.object(drv, '_get_mount_point_for_share') as \
                mock_get_mount_point_for_share,\
                mock.patch.object(drv, '_mount_glusterfs') as \
                mock_mount_glusterfs:
            mock_get_mount_point_for_share.return_value = self.TEST_MNT_POINT
            mock_get_file_mode.return_value = 0o777
            mock_get_file_gid.return_value = 333333

            drv._ensure_share_mounted(self.TEST_EXPORT1)

            mock_get_file_mode.assert_called_once_with(self.TEST_MNT_POINT)
            mock_get_file_gid.assert_called_once_with(self.TEST_MNT_POINT)
            mock_ensure_share_writable.assert_called_once_with(
                self.TEST_MNT_POINT)
            self.assertTrue(mock_ensure_share_writable.called)
            self.assertTrue(mock_mount_glusterfs.called)
            self.assertTrue(mock_execute.called)

    def test_ensure_shares_mounted_should_save_mounting_successfully(self):
        """_ensure_shares_mounted should save share if mounted with success."""
        drv = self._driver

        with mock.patch.object(drv, '_read_config_file') as \
                mock_read_config_file,\
                mock.patch.object(drv, '_ensure_share_mounted') as \
                mock_ensure_share_mounted:
            config_data = []
            config_data.append(self.TEST_EXPORT1)
            mock_read_config_file.return_value = config_data

            drv._ensure_shares_mounted()

            mock_ensure_share_mounted.\
                assert_called_once_with(self.TEST_EXPORT1)
            self.assertEqual(1, len(drv._mounted_shares))
            self.assertEqual(self.TEST_EXPORT1, drv._mounted_shares[0])

    def test_ensure_shares_mounted_should_not_save_mounting_with_error(self):
        """_ensure_shares_mounted should not save share if failed to mount."""
        drv = self._driver

        with mock.patch.object(drv, '_read_config_file') as \
                mock_read_config_file,\
                mock.patch.object(drv, '_ensure_share_mounted') as \
                mock_ensure_share_mounted:
            config_data = []
            config_data.append(self.TEST_EXPORT1)
            mock_read_config_file.return_value = config_data
            mock_ensure_share_mounted.side_effect = Exception()

            drv._ensure_shares_mounted()

            self.assertEqual(0, len(drv._mounted_shares))

    def test_setup_should_throw_error_if_shares_config_not_configured(self):
        """do_setup should throw error if shares config is not configured."""
        drv = self._driver

        drv.configuration.glusterfs_shares_config = None

        self.assertRaisesAndMessageMatches(exception.GlusterfsException,
                                           'no Gluster config file configured',
                                           drv.do_setup,
                                           mock.MagicMock())

    @mock.patch.object(os.path, 'exists')
    def test_setup_should_throw_exception_if_client_is_not_installed(
            self, mock_exists):
        """do_setup should throw exception if client is not installed."""
        drv = self._driver

        self.override_config("glusterfs_shares_config",
                             self.TEST_SHARES_CONFIG_FILE)

        with mock.patch.object(drv, '_execute') as mock_execute:
            mock_exists.return_value = True
            mock_execute.side_effect = OSError(errno.ENOENT,
                                               'No such file or directory')
            self.assertRaisesAndMessageMatches(exception.GlusterfsException,
                                               'mount.glusterfs is not '
                                               'installed',
                                               drv.do_setup,
                                               mock.MagicMock())

    def _fake_load_shares_config(self, config):
        self._driver.shares = {'127.7.7.7:/gluster1': None}

    def _fake_NamedTemporaryFile(self, prefix=None, dir=None):
        raise OSError('Permission denied!')

    @mock.patch.object(os, 'getegid')
    @mock.patch.object(os.path, 'exists')
    def test_setup_set_share_permissions(self, mock_exists, mock_getegid):
        drv = self._driver

        self.override_config("glusterfs_shares_config",
                             self.TEST_SHARES_CONFIG_FILE)

        with mock.patch.object(drv, '_execute') as mock_execute,\
                mock.patch.object(utils, 'get_file_gid') as \
                mock_get_file_gid,\
                mock.patch.object(utils, 'get_file_mode') as \
                mock_get_file_mode,\
                mock.patch.object(tempfile, 'NamedTemporaryFile') as \
                mock_named_temp,\
                mock.patch.object(os_brick.remotefs.remotefs.RemoteFsClient,
                                  'mount') as mock_mount:
            drv._load_shares_config = self._fake_load_shares_config
            mock_named_temp.return_value = self._fake_NamedTemporaryFile
            mock_exists.return_value = True
            mock_get_file_gid.return_value = 33333
            mock_get_file_mode.return_value = 0o000
            mock_getegid.return_value = 888

            drv.do_setup(mock.MagicMock())

            expected = [
                mock.call('mount.glusterfs', check_exit_code=False),
                mock.call('umount',
                          '/mnt/test/8f0473c9ad824b8b6a27264b9cacb005',
                          run_as_root=True),
                mock.call('chgrp', 888,
                          '/mnt/test/8f0473c9ad824b8b6a27264b9cacb005',
                          run_as_root=True),
                mock.call('chmod', 'g+w',
                          '/mnt/test/8f0473c9ad824b8b6a27264b9cacb005',
                          run_as_root=True)]
            self.assertEqual(expected, mock_execute.mock_calls)
            mock_mount.assert_called_once_with('127.7.7.7:/gluster1', [])

    def test_find_share_should_throw_error_if_there_is_no_mounted_shares(self):
        """_find_share should throw error if there is no mounted shares."""
        drv = self._driver

        drv._mounted_shares = []

        self.assertRaises(exception.GlusterfsNoSharesMounted,
                          drv._find_share,
                          self.TEST_SIZE_IN_GB)

    def test_find_share(self):
        """_find_share simple use case."""
        drv = self._driver

        drv._mounted_shares = [self.TEST_EXPORT1, self.TEST_EXPORT2]

        with mock.patch.object(drv, '_get_available_capacity') as \
                mock_get_available_capacity:
            capacity = {self.TEST_EXPORT1: (2 * units.Gi, 5 * units.Gi),
                        self.TEST_EXPORT2: (3 * units.Gi, 10 * units.Gi)}

            def capacity_side_effect(*args, **kwargs):
                return capacity[args[0]]
            mock_get_available_capacity.side_effect = capacity_side_effect

            self.assertEqual(self.TEST_EXPORT2,
                             drv._find_share(self.TEST_SIZE_IN_GB))

    def test_find_share_should_throw_error_if_there_is_no_enough_place(self):
        """_find_share should throw error if there is no share to host vol."""
        drv = self._driver

        drv._mounted_shares = [self.TEST_EXPORT1, self.TEST_EXPORT2]

        with mock.patch.object(drv, '_get_available_capacity') as \
                mock_get_available_capacity:
            capacity = {self.TEST_EXPORT1: (0, 5 * units.Gi),
                        self.TEST_EXPORT2: (0, 10 * units.Gi)}

            def capacity_side_effect(*args, **kwargs):
                return capacity[args[0]]
            mock_get_available_capacity.side_effect = capacity_side_effect

            self.assertRaises(exception.GlusterfsNoSuitableShareFound,
                              drv._find_share,
                              self.TEST_SIZE_IN_GB)

    def _simple_volume(self, id=None):
        volume = DumbVolume()
        volume['provider_location'] = self.TEST_EXPORT1
        if id is None:
            volume['id'] = self.VOLUME_UUID
        else:
            volume['id'] = id
        # volume['name'] mirrors format from db/sqlalchemy/models.py
        volume['name'] = 'volume-%s' % volume['id']
        volume['size'] = 10
        volume['status'] = 'available'

        return volume

    def test_create_thin_volume(self):
        drv = self._driver
        volume = self._simple_volume()

        self._configuration.nas_volume_prov_type = 'thin'

        with mock.patch.object(drv, '_create_qcow2_file') as \
                mock_create_qcow2_file,\
                mock.patch.object(drv, '_set_rw_permissions_for_all') as \
                mock_set_rw_permissions_for_all:
            drv._do_create_volume(volume)

            volume_path = drv.local_path(volume)
            volume_size = volume['size']
            mock_create_qcow2_file.assert_called_once_with(volume_path,
                                                           volume_size)
            mock_set_rw_permissions_for_all.\
                assert_called_once_with(volume_path)

    def test_create_thick_fallocate_volume(self):
        drv = self._driver
        volume = self._simple_volume()

        self._configuration.nas_volume_prov_type = 'thick'

        with mock.patch.object(drv, '_fallocate') as \
                mock_fallocate,\
                mock.patch.object(drv, '_set_rw_permissions_for_all') as \
                mock_set_rw_permissions_for_all:
            drv._do_create_volume(volume)

            volume_path = drv.local_path(volume)
            volume_size = volume['size']
            mock_fallocate.assert_called_once_with(volume_path,
                                                   volume_size)
            mock_set_rw_permissions_for_all.\
                assert_called_once_with(volume_path)

    def test_create_thick_dd_volume(self):
        drv = self._driver
        volume = self._simple_volume()

        self._configuration.nas_volume_prov_type = 'thick'

        with mock.patch.object(drv, '_fallocate') as \
                mock_fallocate,\
                mock.patch.object(drv, '_create_regular_file') as \
                mock_create_regular_file,\
                mock.patch.object(drv, '_set_rw_permissions_for_all') as \
                mock_set_rw_permissions_for_all:
            mock_fallocate.side_effect = putils.ProcessExecutionError(
                stderr='Fallocate: Operation not supported.')
            drv._do_create_volume(volume)

            volume_path = drv.local_path(volume)
            volume_size = volume['size']
            mock_fallocate.assert_called_once_with(volume_path,
                                                   volume_size)
            mock_create_regular_file.assert_called_once_with(volume_path,
                                                             volume_size)
            mock_set_rw_permissions_for_all.\
                assert_called_once_with(volume_path)

    def test_create_volume_should_ensure_glusterfs_mounted(self):
        """create_volume ensures shares provided in config are mounted."""
        drv = self._driver

        with mock.patch.object(drv, '_find_share') as mock_find_share,\
                mock.patch.object(drv, '_do_create_volume') as \
                mock_do_create_volume,\
                mock.patch.object(drv, '_ensure_shares_mounted') as \
                mock_ensure_shares_mounted:
            volume = DumbVolume()
            volume['size'] = self.TEST_SIZE_IN_GB
            volume['id'] = self.VOLUME_UUID
            drv.create_volume(volume)
            self.assertTrue(mock_ensure_shares_mounted.called)
            self.assertTrue(mock_do_create_volume.called)
            self.assertTrue(mock_find_share.called)

    def test_create_volume_should_return_provider_location(self):
        """create_volume should return provider_location with found share."""
        drv = self._driver

        with mock.patch.object(drv, '_find_share') as mock_find_share,\
                mock.patch.object(drv, '_do_create_volume') as \
                mock_do_create_volume,\
                mock.patch.object(drv, '_ensure_shares_mounted') as \
                mock_ensure_shares_mounted:
            mock_find_share.return_value = self.TEST_EXPORT1

            volume = DumbVolume()
            volume['size'] = self.TEST_SIZE_IN_GB
            volume['id'] = self.VOLUME_UUID
            result = drv.create_volume(volume)
            self.assertEqual(self.TEST_EXPORT1, result['provider_location'])
            self.assertTrue(mock_ensure_shares_mounted.called)
            self.assertTrue(mock_do_create_volume.called)

    def test_create_cloned_volume(self):
        drv = self._driver

        with mock.patch.object(drv, '_create_snapshot') as \
                mock_create_snapshot,\
                mock.patch.object(drv, '_delete_snapshot') as \
                mock_delete_snapshot,\
                mock.patch.object(drv, '_copy_volume_from_snapshot') as \
                mock_copy_volume_from_snapshot:
            volume = self._simple_volume()
            src_vref = self._simple_volume()
            src_vref['id'] = '375e32b2-804a-49f2-b282-85d1d5a5b9e1'
            src_vref['name'] = 'volume-%s' % src_vref['id']
            volume_ref = {'id': volume['id'],
                          'name': volume['name'],
                          'status': volume['status'],
                          'provider_location': volume['provider_location'],
                          'size': volume['size']}
            snap_ref = {'volume_name': src_vref['name'],
                        'name': 'clone-snap-%s' % src_vref['id'],
                        'size': src_vref['size'],
                        'volume_size': src_vref['size'],
                        'volume_id': src_vref['id'],
                        'id': 'tmp-snap-%s' % src_vref['id'],
                        'volume': src_vref}
            drv.create_cloned_volume(volume, src_vref)

            mock_create_snapshot.assert_called_once_with(snap_ref)
            mock_copy_volume_from_snapshot.\
                assert_called_once_with(snap_ref, volume_ref, volume['size'])
            self.assertTrue(mock_delete_snapshot.called)

    @mock.patch('oslo_utils.fileutils.delete_if_exists')
    def test_delete_volume(self, mock_delete_if_exists):
        volume = self._simple_volume()
        volume_filename = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s' % (self.TEST_MNT_POINT, volume_filename)
        info_file = volume_path + '.info'

        with mock.patch.object(self._driver, '_ensure_share_mounted') as \
                mock_ensure_share_mounted,\
                mock.patch.object(self._driver, '_local_volume_dir') as \
                mock_local_volume_dir,\
                mock.patch.object(self._driver, 'get_active_image_from_info') as \
                mock_active_image_from_info,\
                mock.patch.object(self._driver, '_execute') as \
                mock_execute,\
                mock.patch.object(self._driver, '_local_path_volume') as \
                mock_local_path_volume,\
                mock.patch.object(self._driver, '_local_path_volume_info') as \
                mock_local_path_volume_info:
            mock_local_volume_dir.return_value = self.TEST_MNT_POINT
            mock_active_image_from_info.return_value = volume_filename
            mock_local_path_volume.return_value = volume_path
            mock_local_path_volume_info.return_value = info_file

            self._driver.delete_volume(volume)

            mock_ensure_share_mounted.assert_called_once_with(
                volume['provider_location'])
            mock_local_volume_dir.assert_called_once_with(volume)
            mock_active_image_from_info.assert_called_once_with(volume)
            mock_execute.assert_called_once_with('rm', '-f', volume_path,
                                                 run_as_root=True)
            mock_local_path_volume_info.assert_called_once_with(volume)
            mock_local_path_volume.assert_called_once_with(volume)
            mock_delete_if_exists.assert_any_call(volume_path)
            mock_delete_if_exists.assert_any_call(info_file)

    def test_refresh_mounts(self):
        with mock.patch.object(self._driver, '_unmount_shares') as \
                mock_unmount_shares,\
                mock.patch.object(self._driver, '_ensure_shares_mounted') as \
                mock_ensure_shares_mounted:
            self._driver._refresh_mounts()

            self.assertTrue(mock_unmount_shares.called)
            self.assertTrue(mock_ensure_shares_mounted.called)

    def test_refresh_mounts_with_excp(self):
        with mock.patch.object(self._driver, '_unmount_shares') as \
                mock_unmount_shares,\
                mock.patch.object(self._driver, '_ensure_shares_mounted') as \
                mock_ensure_shares_mounted,\
                mock.patch.object(glusterfs, 'LOG') as mock_logger:
            mock_stderr = _("umount: <mnt_path>: target is busy")
            mock_unmount_shares.side_effect = \
                putils.ProcessExecutionError(stderr=mock_stderr)

            self._driver._refresh_mounts()

            self.assertTrue(mock_unmount_shares.called)
            self.assertTrue(mock_logger.warning.called)
            self.assertTrue(mock_ensure_shares_mounted.called)

            mock_unmount_shares.reset_mock()
            mock_ensure_shares_mounted.reset_mock()
            mock_logger.reset_mock()
            mock_logger.warning.reset_mock()

            mock_stderr = _("umount: <mnt_path>: some other error")
            mock_unmount_shares.side_effect = \
                putils.ProcessExecutionError(stderr=mock_stderr)

            self.assertRaises(putils.ProcessExecutionError,
                              self._driver._refresh_mounts)

            self.assertTrue(mock_unmount_shares.called)
            self.assertFalse(mock_ensure_shares_mounted.called)

    def test_unmount_shares_with_excp(self):
        self._driver.shares = {'127.7.7.7:/gluster1': None}

        with mock.patch.object(self._driver, '_load_shares_config') as \
                _mock_load_shares_config,\
                mock.patch.object(self._driver, '_do_umount') as \
                mock_do_umount,\
                mock.patch.object(glusterfs, 'LOG') as \
                mock_logger:
            mock_do_umount.side_effect = Exception()

            self._driver._unmount_shares()

            self.assertTrue(mock_do_umount.called)
            self.assertTrue(mock_logger.warning.called)
            self.assertFalse(mock_logger.debug.called)
            self.assertTrue(_mock_load_shares_config.called)

    def test_unmount_shares_1share(self):
        self._driver.shares = {'127.7.7.7:/gluster1': None}

        with mock.patch.object(self._driver, '_load_shares_config') as \
                _mock_load_shares_config,\
                mock.patch.object(self._driver, '_do_umount') as \
                mock_do_umount:
            self._driver._unmount_shares()

            self.assertTrue(mock_do_umount.called)
            mock_do_umount.assert_called_once_with(True,
                                                   '127.7.7.7:/gluster1')
            self.assertTrue(_mock_load_shares_config.called)

    def test_unmount_shares_2share(self):
        self._driver.shares = {'127.7.7.7:/gluster1': None,
                               '127.7.7.8:/gluster2': None}

        with mock.patch.object(self._driver, '_load_shares_config') as \
                _mock_load_shares_config,\
                mock.patch.object(self._driver, '_do_umount') as \
                mock_do_umount:
            self._driver._unmount_shares()

            mock_do_umount.assert_any_call(True,
                                           '127.7.7.7:/gluster1')
            mock_do_umount.assert_any_call(True,
                                           '127.7.7.8:/gluster2')
            self.assertTrue(_mock_load_shares_config.called)

    def test_do_umount(self):
        test_share = '127.7.7.7:/gluster1'
        test_hashpath = '/hashed/mnt/path'

        with mock.patch.object(self._driver, '_get_mount_point_for_share') as \
                mock_get_mntp_share,\
                mock.patch.object(putils, 'execute') as mock_execute:
            mock_get_mntp_share.return_value = test_hashpath

            self._driver._do_umount(True, test_share)

            self.assertTrue(mock_get_mntp_share.called)
            self.assertTrue(mock_execute.called)
            mock_get_mntp_share.assert_called_once_with(test_share)

            cmd = ['umount', test_hashpath]
            self.assertEqual(cmd[0], mock_execute.call_args[0][0])
            self.assertEqual(cmd[1], mock_execute.call_args[0][1])
            self.assertTrue(mock_execute.call_args[1]['run_as_root'])

            mock_get_mntp_share.reset_mock()
            mock_get_mntp_share.return_value = test_hashpath
            mock_execute.reset_mock()

            self._driver._do_umount(False, test_share)

            self.assertTrue(mock_get_mntp_share.called)
            self.assertTrue(mock_execute.called)
            mock_get_mntp_share.assert_called_once_with(test_share)
            cmd = ['umount', test_hashpath]
            self.assertEqual(cmd[0], mock_execute.call_args[0][0])
            self.assertEqual(cmd[1], mock_execute.call_args[0][1])
            self.assertTrue(mock_execute.call_args[1]['run_as_root'])

    def test_do_umount_with_excp1(self):
        test_share = '127.7.7.7:/gluster1'
        test_hashpath = '/hashed/mnt/path'

        with mock.patch.object(self._driver, '_get_mount_point_for_share') as \
                mock_get_mntp_share,\
                mock.patch.object(putils, 'execute') as mock_execute,\
                mock.patch.object(glusterfs, 'LOG') as mock_logger:
            mock_get_mntp_share.return_value = test_hashpath
            mock_execute.side_effect = putils.ProcessExecutionError
            self.assertRaises(putils.ProcessExecutionError,
                              self._driver._do_umount, False,
                              test_share)

            mock_logger.reset_mock()
            mock_logger.info.reset_mock()
            mock_logger.error.reset_mock()
            mock_execute.side_effect = putils.ProcessExecutionError
            try:
                self._driver._do_umount(False, test_share)
            except putils.ProcessExecutionError:
                self.assertFalse(mock_logger.info.called)
                self.assertTrue(mock_logger.error.called)
            except Exception as e:
                self.fail('Unexpected exception thrown:', e)
            else:
                self.fail('putils.ProcessExecutionError not thrown')

    def test_do_umount_with_excp2(self):
        test_share = '127.7.7.7:/gluster1'
        test_hashpath = '/hashed/mnt/path'

        with mock.patch.object(self._driver, '_get_mount_point_for_share') as \
                mock_get_mntp_share,\
                mock.patch.object(putils, 'execute') as mock_execute,\
                mock.patch.object(glusterfs, 'LOG') as mock_logger:
            mock_get_mntp_share.return_value = test_hashpath

            mock_stderr = _("umount: %s: not mounted") % test_hashpath
            mock_execute.side_effect = putils.ProcessExecutionError(
                stderr=mock_stderr)

            self._driver._do_umount(True, test_share)

            self.assertTrue(mock_logger.info.called)
            self.assertFalse(mock_logger.error.called)

            mock_logger.reset_mock()
            mock_logger.info.reset_mock()
            mock_logger.error.reset_mock()
            mock_stderr = _("umount: %s: target is busy") %\
                           (test_hashpath)
            mock_execute.side_effect = putils.ProcessExecutionError(
                stderr=mock_stderr)

            self.assertRaises(putils.ProcessExecutionError,
                              self._driver._do_umount, True,
                              test_share)

            mock_logger.reset_mock()
            mock_logger.info.reset_mock()
            mock_logger.error.reset_mock()
            mock_stderr = _('umount: %s: target is busy') %\
                           (test_hashpath)
            mock_execute.side_effect = putils.ProcessExecutionError(
                stderr=mock_stderr)

            try:
                self._driver._do_umount(True, test_share)
            except putils.ProcessExecutionError:
                self.assertFalse(mock_logger.info.called)
                self.assertTrue(mock_logger.error.called)
            except Exception as e:
                self.fail('Unexpected exception thrown:', e)
            else:
                self.fail('putils.ProcessExecutionError not thrown')

    def test_delete_should_ensure_share_mounted(self):
        """delete_volume should ensure that corresponding share is mounted."""
        drv = self._driver

        with mock.patch.object(drv, '_execute') as mock_execute,\
                mock.patch.object(drv, '_ensure_share_mounted') as \
                mock_ensure_share_mounted:
            volume = DumbVolume()
            volume['id'] = self.VOLUME_UUID
            volume['name'] = 'volume-123'
            volume['provider_location'] = self.TEST_EXPORT1

            drv.delete_volume(volume)

            mock_ensure_share_mounted.\
                assert_called_once_with(self.TEST_EXPORT1)
            self.assertTrue(mock_execute.called)

    def test_delete_should_not_delete_if_provider_location_not_provided(self):
        """delete_volume shouldn't delete if provider_location missed."""
        drv = self._driver

        with mock.patch.object(drv, '_execute') as mock_execute,\
            mock.patch.object(drv, '_ensure_share_mounted') as \
                mock_ensure_share_mounted:
            volume = DumbVolume()
            volume['id'] = self.VOLUME_UUID
            volume['name'] = 'volume-123'
            volume['provider_location'] = None

            drv.delete_volume(volume)

            self.assertFalse(mock_ensure_share_mounted.called)
            self.assertFalse(mock_execute.called)

    def test_read_info_file(self):
        drv = self._driver

        with mock.patch.object(drv, '_read_file') as mock_read_file:
            hashed = drv._get_hash_str(self.TEST_EXPORT1)
            volume_path = '%s/%s/volume-%s' % (self.TEST_MNT_POINT_BASE,
                                               hashed,
                                               self.VOLUME_UUID)
            info_path = '%s%s' % (volume_path, '.info')

            mock_read_file.return_value = '{"%(id)s": "volume-%(id)s"}' %\
                {'id': self.VOLUME_UUID}

            volume = DumbVolume()
            volume['id'] = self.VOLUME_UUID
            volume['name'] = 'volume-%s' % self.VOLUME_UUID

            info = drv._read_info_file(info_path)

            self.assertEqual('volume-%s' % self.VOLUME_UUID,
                             info[self.VOLUME_UUID])

    def test_extend_volume(self):
        drv = self._driver

        volume = self._simple_volume()

        qemu_img_info_output = """image: volume-%s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 473K
        """ % self.VOLUME_UUID
        img_info = imageutils.QemuImgInfo(qemu_img_info_output)

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info,\
                mock.patch.object(self._driver, '_local_volume_dir') as \
                mock_local_volume_dir,\
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info,\
                mock.patch.object(image_utils, 'resize_image') as \
                mock_resize_image:
            mock_get_active_image_from_info.return_value = volume['name']
            mock_local_volume_dir.return_value = self.TEST_MNT_POINT
            mock_qemu_img_info.return_value = img_info

            drv.extend_volume(volume, 3)
            self.assertTrue(mock_resize_image.called)

    def test_extend_volume_with_snapshot(self):
        drv = self._driver

        volume = self._simple_volume()

        snap_file = 'volume-%s.%s' % (self.VOLUME_UUID,
                                      self.SNAP_UUID)

        qemu_img_info_output = """image: %s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 473K
        """ % snap_file

        img_info = imageutils.QemuImgInfo(qemu_img_info_output)

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info,\
                mock.patch.object(self._driver, '_local_volume_dir') as \
                mock_local_volume_dir,\
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info,\
                mock.patch.object(image_utils, 'resize_image') as \
                mock_resize_image:
            mock_get_active_image_from_info.return_value = snap_file
            mock_local_volume_dir.return_value = self.TEST_MNT_POINT
            mock_qemu_img_info.return_value = img_info

            snap_path = '%s/%s' % (self.TEST_MNT_POINT,
                                   snap_file)
            drv.extend_volume(volume, 3)
            mock_resize_image.assert_called_once_with(snap_path, 3)

    def test_create_snapshot_online(self):
        drv = self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt,
                    'status': 'asdf',
                    'progress': 'asdf'}

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)

        with mock.patch.object(drv, '_do_create_snapshot') as \
                mock_do_create_snapshot,\
                mock.patch.object(db, 'snapshot_get') as mock_snapshot_get,\
                mock.patch.object(drv, '_nova') as mock_nova,\
                mock.patch.object(time, 'sleep') as mock_sleep:
            create_info = {'snapshot_id': snap_ref['id'],
                           'type': 'qcow2',
                           'new_file': snap_file}

            snap_ref_progress = snap_ref.copy()
            snap_ref_progress['status'] = 'creating'

            snap_ref_progress_0p = snap_ref_progress.copy()
            snap_ref_progress_0p['progress'] = '0%'

            snap_ref_progress_50p = snap_ref_progress.copy()
            snap_ref_progress_50p['progress'] = '50%'

            snap_ref_progress_90p = snap_ref_progress.copy()
            snap_ref_progress_90p['progress'] = '90%'

            mock_snapshot_get.side_effect = [
                snap_ref_progress_0p, snap_ref_progress_50p,
                snap_ref_progress_90p
            ]

            drv._create_snapshot_online(snap_ref, snap_file, snap_path)
            mock_do_create_snapshot.\
                assert_called_once_with(snap_ref, snap_file, snap_path)
            mock_nova.create_volume_snapshot.\
                assert_called_once_with(ctxt, self.VOLUME_UUID, create_info)
            self.assertTrue(mock_sleep.called)

    def test_create_snapshot_online_novafailure(self):
        drv = self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt}

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)

        with mock.patch.object(drv, '_do_create_snapshot') as mock_do_create_snapshot,\
                mock.patch.object(db, 'snapshot_get') as mock_snapshot_get,\
                mock.patch.object(drv, '_nova') as mock_nova,\
                mock.patch.object(time, 'sleep') as mock_sleep:
            snap_ref_progress = snap_ref.copy()
            snap_ref_progress['status'] = 'creating'

            snap_ref_progress_0p = snap_ref_progress.copy()
            snap_ref_progress_0p['progress'] = '0%'

            snap_ref_progress_50p = snap_ref_progress.copy()
            snap_ref_progress_50p['progress'] = '50%'

            snap_ref_progress_99p = snap_ref_progress.copy()
            snap_ref_progress_99p['progress'] = '99%'
            snap_ref_progress_99p['status'] = 'error'

            mock_snapshot_get.side_effect = [
                snap_ref_progress_0p, snap_ref_progress_50p,
                snap_ref_progress_99p
            ]

            self.assertRaisesAndMessageMatches(
                exception.RemoteFSException,
                'Nova returned "error" status while creating snapshot.',
                drv._create_snapshot_online,
                snap_ref, snap_file, snap_path)
            self.assertTrue(mock_sleep.called)
            self.assertTrue(mock_nova.create_volume_snapshot.called)
            self.assertTrue(mock_do_create_snapshot.called)

    def test_delete_snapshot_online_1(self):
        """Delete the newest snapshot, with only one snap present."""
        drv = self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap to delete (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt}

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_dir = os.path.join(self.TEST_MNT_POINT_BASE, hashed)
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)
        info_path = '%s.info' % volume_path

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)

        with mock.patch.object(drv, '_execute') as mock_execute,\
                mock.patch.object(db, 'snapshot_get') as mock_snapshot_get,\
                mock.patch.object(drv, '_nova') as mock_nova,\
                mock.patch.object(time, 'sleep') as mock_sleep,\
                mock.patch.object(drv, '_read_info_file') as \
                mock_read_info_file,\
                mock.patch.object(drv, '_write_info_file') as \
                mock_write_info_file,\
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info,\
                mock.patch.object(drv, '_ensure_share_writable') as \
                mock_ensure_share_writable:
            snap_info = {'active': snap_file,
                         self.SNAP_UUID: snap_file}
            mock_read_info_file.return_value = snap_info

            qemu_img_info_output = """image: %s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            backing file: %s
            """ % (snap_file, volume_file)
            img_info = imageutils.QemuImgInfo(qemu_img_info_output)

            vol_qemu_img_info_output = """image: %s
            file format: raw
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            """ % volume_file
            volume_img_info = imageutils.QemuImgInfo(vol_qemu_img_info_output)

            paths = {snap_path: img_info, volume_path: volume_img_info}

            def img_info_side_effect(*args, **kwargs):
                return paths[args[0]]

            mock_qemu_img_info.side_effect = img_info_side_effect

            delete_info = {
                'type': 'qcow2',
                'merge_target_file': None,
                'file_to_merge': None,
                'volume_id': self.VOLUME_UUID
            }

            snap_ref_progress = snap_ref.copy()
            snap_ref_progress['status'] = 'deleting'

            snap_ref_progress_0p = snap_ref_progress.copy()
            snap_ref_progress_0p['progress'] = '0%'

            snap_ref_progress_50p = snap_ref_progress.copy()
            snap_ref_progress_50p['progress'] = '50%'

            snap_ref_progress_90p = snap_ref_progress.copy()
            snap_ref_progress_90p['progress'] = '90%'

            mock_snapshot_get.side_effect = [
                snap_ref_progress_0p, snap_ref_progress_50p,
                snap_ref_progress_90p
            ]

            drv.delete_snapshot(snap_ref)

            mock_ensure_share_writable.assert_called_once_with(volume_dir)
            mock_nova.delete_volume_snapshot.\
                assert_called_once_with(ctxt, self.SNAP_UUID, delete_info)
            mock_write_info_file.assert_called_once_with(info_path, snap_info)
            mock_execute.assert_called_once_with('rm', '-f', volume_path,
                                                 run_as_root=True)
            self.assertTrue(mock_ensure_share_writable.called)
            self.assertTrue(mock_write_info_file.called)
            self.assertTrue(mock_sleep.called)
            self.assertTrue(mock_nova.delete_volume_snapshot.called)
            self.assertTrue(mock_execute.called)

    def test_delete_snapshot_online_2(self):
        """Delete the middle of 3 snapshots."""
        drv = self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap to delete (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt}

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_dir = os.path.join(self.TEST_MNT_POINT_BASE, hashed)
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)
        info_path = '%s.info' % volume_path

        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)
        snap_file_2 = '%s.%s' % (volume_file, self.SNAP_UUID_2)

        with mock.patch.object(drv, '_execute') as mock_execute,\
                mock.patch.object(db, 'snapshot_get') as \
                mock_snapshot_get,\
                mock.patch.object(drv, '_nova') as \
                mock_nova,\
                mock.patch.object(time, 'sleep') as \
                mock_sleep,\
                mock.patch.object(drv, '_read_info_file') as \
                mock_read_info_file,\
                mock.patch.object(drv, '_write_info_file') as \
                mock_write_info_file,\
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info,\
                mock.patch.object(drv, '_ensure_share_writable') as \
                mock_ensure_share_writable:
            snap_info = {'active': snap_file_2,
                         self.SNAP_UUID: snap_file,
                         self.SNAP_UUID_2: snap_file_2}

            mock_read_info_file.return_value = snap_info

            qemu_img_info_output = """image: %s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            backing file: %s
            """ % (snap_file, volume_file)
            img_info = imageutils.QemuImgInfo(qemu_img_info_output)

            vol_qemu_img_info_output = """image: %s
            file format: raw
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            """ % volume_file
            volume_img_info = imageutils.QemuImgInfo(vol_qemu_img_info_output)

            paths = {snap_path: img_info, volume_path: volume_img_info}

            def img_info_side_effect(*args, **kwargs):
                return paths[args[0]]
            mock_qemu_img_info.side_effect = img_info_side_effect

            delete_info = {'type': 'qcow2',
                           'merge_target_file': volume_file,
                           'file_to_merge': snap_file,
                           'volume_id': self.VOLUME_UUID}

            snap_ref_progress = snap_ref.copy()
            snap_ref_progress['status'] = 'deleting'

            snap_ref_progress_0p = snap_ref_progress.copy()
            snap_ref_progress_0p['progress'] = '0%'

            snap_ref_progress_50p = snap_ref_progress.copy()
            snap_ref_progress_50p['progress'] = '50%'

            snap_ref_progress_90p = snap_ref_progress.copy()
            snap_ref_progress_90p['progress'] = '90%'

            mock_snapshot_get.side_effect = [
                snap_ref_progress_0p, snap_ref_progress_50p,
                snap_ref_progress_90p]

            drv.delete_snapshot(snap_ref)

            mock_ensure_share_writable.assert_called_once_with(volume_dir)
            mock_nova.delete_volume_snapshot.\
                assert_called_once_with(ctxt, self.SNAP_UUID, delete_info)
            mock_write_info_file.assert_called_once_with(info_path, snap_info)
            mock_execute.assert_called_once_with('rm', '-f',
                                                 snap_path, run_as_root=True)
            self.assertTrue(mock_ensure_share_writable.called)
            self.assertTrue(mock_write_info_file.called)
            self.assertTrue(mock_sleep.called)
            self.assertTrue(mock_nova.delete_volume_snapshot.called)
            self.assertTrue(mock_execute.called)

    def test_delete_snapshot_online_novafailure(self):
        """Delete the newest snapshot."""
        drv = self._driver

        volume = self._simple_volume()
        volume['status'] = 'in-use'

        ctxt = context.RequestContext('fake_user', 'fake_project')

        snap_ref = {'name': 'test snap to delete (online)',
                    'volume_id': self.VOLUME_UUID,
                    'volume': volume,
                    'id': self.SNAP_UUID,
                    'context': ctxt}

        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        volume_file = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s/%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    volume_file)
        snap_path = '%s.%s' % (volume_path, self.SNAP_UUID)
        snap_file = '%s.%s' % (volume_file, self.SNAP_UUID)

        with mock.patch.object(drv, '_execute') as mock_execute,\
                mock.patch.object(drv, '_do_create_snapshot') as \
                mock_do_create_snapshot,\
                mock.patch.object(db, 'snapshot_get') as \
                mock_snapshot_get,\
                mock.patch.object(drv, '_nova') as \
                mock_nova,\
                mock.patch.object(time, 'sleep') as \
                mock_sleep,\
                mock.patch.object(drv, '_read_info_file') as \
                mock_read_info_file,\
                mock.patch.object(drv, '_write_info_file') as \
                mock_write_info_file,\
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info,\
                mock.patch.object(drv, '_ensure_share_writable') as \
                mock_ensure_share_writable:
            snap_info = {'active': snap_file,
                         self.SNAP_UUID: snap_file}
            mock_read_info_file.return_value = snap_info

            qemu_img_info_output = """image: %s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            backing file: %s
            """ % (snap_file, volume_file)
            img_info = imageutils.QemuImgInfo(qemu_img_info_output)

            vol_qemu_img_info_output = """image: %s
            file format: raw
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            """ % volume_file
            volume_img_info = imageutils.QemuImgInfo(vol_qemu_img_info_output)

            paths = {snap_path: img_info, volume_path: volume_img_info}

            def img_info_side_effect(*args, **kwargs):
                return paths[args[0]]

            mock_qemu_img_info.side_effect = img_info_side_effect

            snap_ref_progress = snap_ref.copy()
            snap_ref_progress['status'] = 'deleting'

            snap_ref_progress_0p = snap_ref_progress.copy()
            snap_ref_progress_0p['progress'] = '0%'

            snap_ref_progress_50p = snap_ref_progress.copy()
            snap_ref_progress_50p['progress'] = '50%'

            snap_ref_progress_90p = snap_ref_progress.copy()
            snap_ref_progress_90p['status'] = 'error_deleting'
            snap_ref_progress_90p['progress'] = '90%'

            mock_snapshot_get.side_effect = [
                snap_ref_progress_0p, snap_ref_progress_50p,
                snap_ref_progress_90p]
            self.assertRaisesAndMessageMatches(exception.RemoteFSException,
                                               'Unable to delete snapshot',
                                               drv.delete_snapshot,
                                               snap_ref)
            self.assertTrue(mock_ensure_share_writable.called)
            self.assertFalse(mock_write_info_file.called)
            self.assertTrue(mock_sleep.called)
            self.assertFalse(mock_nova.called)
            self.assertFalse(mock_do_create_snapshot.called)
            self.assertFalse(mock_execute.called)

    def test_get_backing_chain_for_path(self):
        drv = self._driver

        self.override_config('glusterfs_mount_point_base',
                             self.TEST_MNT_POINT_BASE)

        volume = self._simple_volume()
        vol_filename = volume['name']
        vol_filename_2 = volume['name'] + '.abcd'
        vol_filename_3 = volume['name'] + '.efef'
        hashed = drv._get_hash_str(self.TEST_EXPORT1)
        vol_dir = '%s/%s' % (self.TEST_MNT_POINT_BASE, hashed)
        vol_path = '%s/%s' % (vol_dir, vol_filename)
        vol_path_2 = '%s/%s' % (vol_dir, vol_filename_2)
        vol_path_3 = '%s/%s' % (vol_dir, vol_filename_3)

        with mock.patch.object(drv, '_local_volume_dir') as \
                mock_local_volume_dir,\
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info:
            qemu_img_output_base = """image: %(image_name)s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            """
            qemu_img_output = """image: %(image_name)s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            backing file: %(backing_file)s
            """

            qemu_img_output_1 = qemu_img_output_base %\
                {'image_name': vol_filename}
            qemu_img_output_2 = qemu_img_output %\
                {'image_name': vol_filename_2,
                 'backing_file': vol_filename}
            qemu_img_output_3 = qemu_img_output %\
                {'image_name': vol_filename_3,
                 'backing_file': vol_filename_2}

            info_1 = imageutils.QemuImgInfo(qemu_img_output_1)
            info_2 = imageutils.QemuImgInfo(qemu_img_output_2)
            info_3 = imageutils.QemuImgInfo(qemu_img_output_3)

            img_infos = {vol_path_3: info_3,
                         vol_path_2: info_2,
                         vol_path: info_1}

            def img_info_side_effect(*args, **kwargs):
                return img_infos[args[0]]

            mock_qemu_img_info.side_effect = img_info_side_effect
            mock_local_volume_dir.return_value = vol_dir

            chain = drv._get_backing_chain_for_path(volume, vol_path_3)

            # Verify chain contains all expected data
            item_1 = drv._get_matching_backing_file(chain, vol_filename)
            self.assertEqual(vol_filename_2, item_1['filename'])
            chain.remove(item_1)
            item_2 = drv._get_matching_backing_file(chain, vol_filename_2)
            self.assertEqual(vol_filename_3, item_2['filename'])
            chain.remove(item_2)
            self.assertEqual(1, len(chain))
            self.assertEqual(vol_filename, chain[0]['filename'])

    def test_copy_volume_from_snapshot(self):
        drv = self._driver

        with mock.patch.object(image_utils, 'convert_image') as \
                mock_convert_image,\
                mock.patch.object(drv, '_read_info_file') as \
                mock_read_info_file,\
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info,\
                mock.patch.object(drv, '_set_rw_permissions_for_all') as \
                mock_set_rw_permissions:
            dest_volume = self._simple_volume(
                'c1073000-0000-0000-0000-0000000c1073')
            src_volume = self._simple_volume()

            vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                                   drv._get_hash_str(self.TEST_EXPORT1))
            src_vol_path = os.path.join(vol_dir, src_volume['name'])
            dest_vol_path = os.path.join(vol_dir, dest_volume['name'])
            snapshot = {'volume_name': src_volume['name'],
                        'name': 'clone-snap-%s' % src_volume['id'],
                        'size': src_volume['size'],
                        'volume_size': src_volume['size'],
                        'volume_id': src_volume['id'],
                        'id': 'tmp-snap-%s' % src_volume['id'],
                        'volume': src_volume}
            snap_file = dest_volume['name'] + '.' + snapshot['id']
            size = dest_volume['size']
            mock_read_info_file.return_value = {'active': snap_file,
                                                snapshot['id']: snap_file}
            qemu_img_output = """image: %s
            file format: raw
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            backing file: %s
            """ % (snap_file, src_volume['name'])
            img_info = imageutils.QemuImgInfo(qemu_img_output)
            mock_qemu_img_info.return_value = img_info

            drv._copy_volume_from_snapshot(snapshot, dest_volume, size)

            mock_convert_image.assert_called_once_with(src_vol_path,
                                                       dest_vol_path, 'raw')
            mock_set_rw_permissions.assert_called_once_with(dest_vol_path)

    def test_create_volume_from_snapshot(self):
        drv = self._driver

        src_volume = self._simple_volume()
        snap_ref = {'volume_name': src_volume['name'],
                    'name': 'clone-snap-%s' % src_volume['id'],
                    'size': src_volume['size'],
                    'volume_size': src_volume['size'],
                    'volume_id': src_volume['id'],
                    'id': 'tmp-snap-%s' % src_volume['id'],
                    'volume': src_volume,
                    'status': 'available'}

        new_volume = DumbVolume()
        new_volume['id'] = self.VOLUME_UUID
        new_volume['size'] = snap_ref['size']

        with mock.patch.object(drv, '_ensure_shares_mounted') as \
                mock_ensure_shares_mounted,\
                mock.patch.object(drv, '_find_share') as \
                mock_find_share, \
                mock.patch.object(drv, '_do_create_volume') as \
                mock_do_create_volume, \
                mock.patch.object(drv, '_copy_volume_from_snapshot') as \
                mock_copy_volume:
            mock_find_share.return_value = self.TEST_EXPORT1
            drv.create_volume_from_snapshot(new_volume, snap_ref)

            self.assertTrue(mock_ensure_shares_mounted.called)
            mock_do_create_volume.assert_called_once_with(new_volume)
            mock_copy_volume.assert_called_once_with(snap_ref,
                                                     new_volume,
                                                     new_volume['size'])

    def test_initialize_connection(self):
        drv = self._driver

        volume = self._simple_volume()
        qemu_img_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        """ % volume['name']
        img_info = imageutils.QemuImgInfo(qemu_img_output)

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info,\
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info:
            mock_get_active_image_from_info.return_value = volume['name']
            mock_qemu_img_info.return_value = img_info

            conn_info = drv.initialize_connection(volume, None)

            self.assertEqual('raw', conn_info['data']['format'])
            self.assertEqual('glusterfs', conn_info['driver_volume_type'])
            self.assertEqual(volume['name'], conn_info['data']['name'])
            self.assertEqual(self.TEST_MNT_POINT_BASE,
                             conn_info['mount_point_base'])

    def test_get_mount_point_base(self):
        drv = self._driver

        self.assertEqual(self.TEST_MNT_POINT_BASE,
                         drv._get_mount_point_base())

    def test_backup_volume(self):
        """Backup a volume with no snapshots."""
        drv = self._driver

        with mock.patch.object(drv.db, 'volume_get') as mock_volume_get,\
                mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info,\
                mock.patch.object(drv, '_qemu_img_info') as \
                mock_qemu_img_info,\
                mock.patch.object(base_driver.BaseVD, 'backup_volume') as \
                mock_backup_volume:
            ctxt = context.RequestContext('fake_user', 'fake_project')
            volume = self._simple_volume()
            backup = {'volume_id': volume['id']}
            mock_volume_get.return_value = volume
            mock_get_active_image_from_info.return_value = '/some/path'

            info = imageutils.QemuImgInfo()
            info.file_format = 'raw'
            mock_qemu_img_info.return_value = info

            drv.backup_volume(ctxt, backup, mock.MagicMock())
            self.assertTrue(mock_backup_volume.called)

    def test_backup_volume_previous_snap(self):
        """Backup a volume that previously had a snapshot.

           Snapshot was deleted, snap_info is different from above.
        """
        drv = self._driver

        with mock.patch.object(drv.db, 'volume_get') as mock_volume_get,\
                mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info,\
                mock.patch.object(drv, '_qemu_img_info') as \
                mock_qemu_img_info,\
                mock.patch.object(base_driver.BaseVD, 'backup_volume') as \
                mock_backup_volume:
            ctxt = context.RequestContext('fake_user', 'fake_project')
            volume = self._simple_volume()
            backup = {'volume_id': volume['id']}
            mock_volume_get.return_value = volume
            mock_get_active_image_from_info.return_value = '/some/file2'

            info = imageutils.QemuImgInfo()
            info.file_format = 'raw'
            mock_qemu_img_info.return_value = info

            drv.backup_volume(ctxt, backup, mock.MagicMock())
            self.assertTrue(mock_backup_volume.called)

    def test_backup_snap_failure_1(self):
        """Backup fails if snapshot exists (database)."""

        drv = self._driver

        with mock.patch.object(drv.db, 'snapshot_get_all_for_volume') as \
                mock_snapshot_get_all_for_volume:
            ctxt = context.RequestContext('fake_user', 'fake_project')
            volume = self._simple_volume()
            backup = {'volume_id': volume['id']}
            mock_snapshot_get_all_for_volume.return_value = [
                {'snap1': 'a'},
                {'snap2': 'b'}
            ]
            self.assertRaises(exception.InvalidVolume,
                              drv.backup_volume,
                              ctxt, backup, mock.MagicMock())

    def test_backup_snap_failure_2(self):
        """Backup fails if snapshot exists (on-disk)."""
        drv = self._driver

        with mock.patch.object(drv.db, 'volume_get') as mock_volume_get,\
                mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info, \
                mock.patch.object(drv, '_qemu_img_info') as \
                mock_qemu_img_info:
            ctxt = context.RequestContext('fake_user', 'fake_project')
            volume = self._simple_volume()
            backup = {'volume_id': volume['id']}
            mock_volume_get.return_value = volume
            mock_get_active_image_from_info.return_value = '/some/path/file2'

            info = imageutils.QemuImgInfo()
            info.file_format = 'raw'
            info.backing_file = 'file1'
            mock_qemu_img_info.return_value = info

            self.assertRaises(exception.InvalidVolume,
                              drv.backup_volume,
                              ctxt, backup, mock.MagicMock())

    def test_backup_failure_unsupported_format(self):
        """Attempt to backup a volume with a qcow2 base."""
        drv = self._driver

        with mock.patch.object(drv.db, 'volume_get') as mock_volume_get,\
                mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info,\
                mock.patch.object(drv, '_qemu_img_info') as mock_qemu_img_info:
            ctxt = context.RequestContext('fake_user', 'fake_project')
            volume = self._simple_volume()
            backup = {'volume_id': volume['id']}
            mock_volume_get.return_value = volume
            mock_get_active_image_from_info.return_value = '/some/path'

            info = imageutils.QemuImgInfo()
            info.file_format = 'qcow2'

            self.assertRaises(exception.InvalidVolume,
                              drv.backup_volume,
                              ctxt, backup, mock.MagicMock())

            mock_volume_get.return_value = volume
            mock_qemu_img_info.return_value = info

            self.assertRaises(exception.InvalidVolume,
                              drv.backup_volume,
                              ctxt, backup, mock.MagicMock())

    def test_copy_volume_to_image_raw_image(self):
        drv = self._driver

        volume = self._simple_volume()
        volume_path = '%s/%s' % (self.TEST_MNT_POINT, volume['name'])
        image_meta = {'id': '10958016-e196-42e3-9e7f-5d8927ae3099'}

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info, \
                mock.patch.object(drv, '_local_volume_dir') as \
                mock_local_volume_dir, \
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info, \
                mock.patch.object(image_utils, 'upload_volume') as \
                mock_upload_volume, \
                mock.patch.object(image_utils, 'create_temporary_file') as \
                mock_create_temporary_file:
            mock_get_active_image_from_info.return_value = volume['name']

            mock_local_volume_dir.return_value = self.TEST_MNT_POINT

            mock_create_temporary_file.return_value = self.TEST_TMP_FILE

            qemu_img_output = """image: %s
            file format: raw
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            """ % volume['name']
            img_info = imageutils.QemuImgInfo(qemu_img_output)
            mock_qemu_img_info.return_value = img_info

            upload_path = volume_path

            drv.copy_volume_to_image(mock.ANY, volume, mock.ANY, image_meta)

            mock_get_active_image_from_info.assert_called_once_with(volume)
            mock_local_volume_dir.assert_called_once_with(volume)
            mock_qemu_img_info.assert_called_once_with(volume_path)
            mock_upload_volume.assert_called_once_with(
                mock.ANY, mock.ANY, mock.ANY, upload_path)
            self.assertEqual(1, mock_create_temporary_file.call_count)

    def test_copy_volume_to_image_qcow2_image(self):
        """Upload a qcow2 image file which has to be converted to raw first."""
        drv = self._driver

        volume = self._simple_volume()
        volume_path = '%s/%s' % (self.TEST_MNT_POINT, volume['name'])
        image_meta = {'id': '10958016-e196-42e3-9e7f-5d8927ae3099'}

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info, \
                mock.patch.object(drv, '_local_volume_dir') as \
                mock_local_volume_dir, \
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info, \
                mock.patch.object(image_utils, 'convert_image') as \
                mock_convert_image, \
                mock.patch.object(image_utils, 'upload_volume') as \
                mock_upload_volume, \
                mock.patch.object(image_utils, 'create_temporary_file') as \
                mock_create_temporary_file:
            mock_get_active_image_from_info.return_value = volume['name']

            mock_local_volume_dir.return_value = self.TEST_MNT_POINT

            mock_create_temporary_file.return_value = self.TEST_TMP_FILE

            qemu_img_output = """image: %s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            """ % volume['name']
            img_info = imageutils.QemuImgInfo(qemu_img_output)
            mock_qemu_img_info.return_value = img_info

            upload_path = self.TEST_TMP_FILE

            drv.copy_volume_to_image(mock.ANY, volume, mock.ANY, image_meta)

            mock_get_active_image_from_info.assert_called_once_with(volume)
            mock_local_volume_dir.assert_called_with(volume)
            mock_qemu_img_info.assert_called_once_with(volume_path)
            mock_convert_image.assert_called_once_with(
                volume_path, upload_path, 'raw')
            mock_upload_volume.assert_called_once_with(
                mock.ANY, mock.ANY, mock.ANY, upload_path)
            self.assertEqual(1, mock_create_temporary_file.call_count)

    def test_copy_volume_to_image_snapshot_exists(self):
        """Upload an active snapshot which has to be converted to raw first."""
        drv = self._driver

        volume = self._simple_volume()
        volume_path = '%s/volume-%s' % (self.TEST_MNT_POINT, self.VOLUME_UUID)
        volume_filename = 'volume-%s' % self.VOLUME_UUID
        image_meta = {'id': '10958016-e196-42e3-9e7f-5d8927ae3099'}

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info, \
                mock.patch.object(drv, '_local_volume_dir') as \
                mock_local_volume_dir, \
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info, \
                mock.patch.object(image_utils, 'convert_image') as \
                mock_convert_image, \
                mock.patch.object(image_utils, 'upload_volume') as \
                mock_upload_volume, \
                mock.patch.object(image_utils, 'create_temporary_file') as \
                mock_create_temporary_file:
            mock_get_active_image_from_info.return_value = volume['name']

            mock_local_volume_dir.return_value = self.TEST_MNT_POINT

            mock_create_temporary_file.return_value = self.TEST_TMP_FILE

            qemu_img_output = """image: volume-%s.%s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            backing file: %s
            """ % (self.VOLUME_UUID, self.SNAP_UUID, volume_filename)
            img_info = imageutils.QemuImgInfo(qemu_img_output)
            mock_qemu_img_info.return_value = img_info

            upload_path = self.TEST_TMP_FILE

            drv.copy_volume_to_image(mock.ANY, volume, mock.ANY, image_meta)

            mock_get_active_image_from_info.assert_called_once_with(volume)
            mock_local_volume_dir.assert_called_with(volume)
            mock_qemu_img_info.assert_called_once_with(volume_path)
            mock_convert_image.assert_called_once_with(
                volume_path, upload_path, 'raw')
            mock_upload_volume.assert_called_once_with(
                mock.ANY, mock.ANY, mock.ANY, upload_path)
            self.assertEqual(1, mock_create_temporary_file.call_count)

    def test_migrate_volume_is_there(self):
        """Ensure that driver.migrate_volume() is there."""

        drv = self._driver

        ctxt = context.RequestContext('fake_user', 'fake_project')
        volume = self._simple_volume()
        ret = drv.migrate_volume(ctxt,
                                 volume,
                                 mock.sentinel.host)

        self.assertEqual((False, None), ret)

    def test_manage_existing_is_there(self):
        """Ensure that driver.manage_existing() is there."""

        drv = self._driver

        volume = self._simple_volume(id=mock.sentinel.manage_id)

        self.assertRaises(NotImplementedError,
                          drv.manage_existing,
                          volume, mock.sentinel.existing_ref)

    def test_unmanage_is_there(self):
        """Ensure that driver.unmanage() is there."""

        drv = self._driver

        volume = self._simple_volume(id=mock.sentinel.unmanage_id)
        self.assertRaises(NotImplementedError,
                          drv.unmanage,
                          volume)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Scheduler
"""

import mock
from oslo_config import cfg

from cinder import context
from cinder import db
from cinder import exception
from cinder.objects import fields
from cinder.scheduler import driver
from cinder.scheduler import filter_scheduler
from cinder.scheduler import manager
from cinder import test
from cinder.tests.unit import fake_consistencygroup
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit import utils as tests_utils

CONF = cfg.CONF


class SchedulerManagerTestCase(test.TestCase):
    """Test case for scheduler manager."""

    manager_cls = manager.SchedulerManager
    driver_cls = driver.Scheduler
    driver_cls_name = 'cinder.scheduler.driver.Scheduler'

    class AnException(Exception):
        pass

    def setUp(self):
        super(SchedulerManagerTestCase, self).setUp()
        self.flags(scheduler_driver=self.driver_cls_name)
        self.manager = self.manager_cls()
        self.manager._startup_delay = False
        self.context = context.get_admin_context()
        self.topic = 'fake_topic'
        self.fake_args = (1, 2, 3)
        self.fake_kwargs = {'cat': 'meow', 'dog': 'woof'}

    def test_1_correct_init(self):
        # Correct scheduler driver
        manager = self.manager
        self.assertIsInstance(manager.driver, self.driver_cls)

    @mock.patch('eventlet.sleep')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.publish_service_capabilities')
    def test_init_host_with_rpc(self, publish_capabilities_mock, sleep_mock):
        self.manager._startup_delay = True
        self.manager.init_host_with_rpc()
        publish_capabilities_mock.assert_called_once_with(mock.ANY)
        sleep_mock.assert_called_once_with(CONF.periodic_interval)
        self.assertFalse(self.manager._startup_delay)

    @mock.patch('cinder.rpc.LAST_RPC_VERSIONS', {'cinder-volume': '1.3'})
    @mock.patch('cinder.rpc.LAST_OBJ_VERSIONS', {'cinder-volume': '1.5'})
    def test_reset(self):
        mgr = self.manager_cls()

        volume_rpcapi = mgr.driver.volume_rpcapi
        self.assertEqual('1.3', volume_rpcapi.client.version_cap)
        self.assertEqual('1.5',
                         volume_rpcapi.client.serializer._base.version_cap)
        mgr.reset()

        volume_rpcapi = mgr.driver.volume_rpcapi
        self.assertIsNone(volume_rpcapi.client.version_cap)
        self.assertIsNone(volume_rpcapi.client.serializer._base.version_cap)

    @mock.patch('cinder.scheduler.driver.Scheduler.'
                'update_service_capabilities')
    def test_update_service_capabilities_empty_dict(self, _mock_update_cap):
        # Test no capabilities passes empty dictionary
        service = 'fake_service'
        host = 'fake_host'

        self.manager.update_service_capabilities(self.context,
                                                 service_name=service,
                                                 host=host)
        _mock_update_cap.assert_called_once_with(service, host, {})

    @mock.patch('cinder.scheduler.driver.Scheduler.'
                'update_service_capabilities')
    def test_update_service_capabilities_correct(self, _mock_update_cap):
        # Test capabilities passes correctly
        service = 'fake_service'
        host = 'fake_host'
        capabilities = {'fake_capability': 'fake_value'}

        self.manager.update_service_capabilities(self.context,
                                                 service_name=service,
                                                 host=host,
                                                 capabilities=capabilities)
        _mock_update_cap.assert_called_once_with(service, host, capabilities)

    @mock.patch('cinder.scheduler.driver.Scheduler.schedule_create_volume')
    @mock.patch('cinder.db.volume_update')
    def test_create_volume_exception_puts_volume_in_error_state(
            self, _mock_volume_update, _mock_sched_create):
        # Test NoValidHost exception behavior for create_volume.
        # Puts the volume in 'error' state and eats the exception.
        _mock_sched_create.side_effect = exception.NoValidHost(reason="")
        volume = fake_volume.fake_volume_obj(self.context)
        topic = 'fake_topic'
        request_spec = {'volume_id': volume.id}

        self.manager.create_volume(self.context, topic, volume.id,
                                   request_spec=request_spec,
                                   filter_properties={},
                                   volume=volume)
        _mock_volume_update.assert_called_once_with(self.context,
                                                    volume.id,
                                                    {'status': 'error'})
        _mock_sched_create.assert_called_once_with(self.context, request_spec,
                                                   {})

    @mock.patch('cinder.scheduler.driver.Scheduler.schedule_create_volume')
    @mock.patch('eventlet.sleep')
    def test_create_volume_no_delay(self, _mock_sleep, _mock_sched_create):
        volume = fake_volume.fake_volume_obj(self.context)
        topic = 'fake_topic'

        request_spec = {'volume_id': volume.id}

        self.manager.create_volume(self.context, topic, volume.id,
                                   request_spec=request_spec,
                                   filter_properties={},
                                   volume=volume)
        _mock_sched_create.assert_called_once_with(self.context, request_spec,
                                                   {})
        self.assertFalse(_mock_sleep.called)

    @mock.patch('cinder.scheduler.driver.Scheduler.schedule_create_volume')
    @mock.patch('cinder.scheduler.driver.Scheduler.is_ready')
    @mock.patch('eventlet.sleep')
    def test_create_volume_delay_scheduled_after_3_tries(self, _mock_sleep,
                                                         _mock_is_ready,
                                                         _mock_sched_create):
        self.manager._startup_delay = True
        volume = fake_volume.fake_volume_obj(self.context)
        topic = 'fake_topic'

        request_spec = {'volume_id': volume.id}

        _mock_is_ready.side_effect = [False, False, True]

        self.manager.create_volume(self.context, topic, volume.id,
                                   request_spec=request_spec,
                                   filter_properties={},
                                   volume=volume)
        _mock_sched_create.assert_called_once_with(self.context, request_spec,
                                                   {})
        calls = [mock.call(1)] * 2
        _mock_sleep.assert_has_calls(calls)
        self.assertEqual(2, _mock_sleep.call_count)

    @mock.patch('cinder.scheduler.driver.Scheduler.schedule_create_volume')
    @mock.patch('cinder.scheduler.driver.Scheduler.is_ready')
    @mock.patch('eventlet.sleep')
    def test_create_volume_delay_scheduled_in_1_try(self, _mock_sleep,
                                                    _mock_is_ready,
                                                    _mock_sched_create):
        self.manager._startup_delay = True
        volume = fake_volume.fake_volume_obj(self.context)
        topic = 'fake_topic'

        request_spec = {'volume_id': volume.id}

        _mock_is_ready.return_value = True

        self.manager.create_volume(self.context, topic, volume.id,
                                   request_spec=request_spec,
                                   filter_properties={},
                                   volume=volume)
        _mock_sched_create.assert_called_once_with(self.context, request_spec,
                                                   {})
        self.assertFalse(_mock_sleep.called)

    @mock.patch('cinder.db.volume_get')
    @mock.patch('cinder.scheduler.driver.Scheduler.host_passes_filters')
    @mock.patch('cinder.db.volume_update')
    def test_migrate_volume_exception_returns_volume_state(
            self, _mock_volume_update, _mock_host_passes,
            _mock_volume_get):
        # Test NoValidHost exception behavior for migrate_volume_to_host.
        # Puts the volume in 'error_migrating' state and eats the exception.
        fake_updates = {'migration_status': 'error'}
        self._test_migrate_volume_exception_returns_volume_state(
            _mock_volume_update, _mock_host_passes, _mock_volume_get,
            'available', fake_updates)

    @mock.patch('cinder.db.volume_get')
    @mock.patch('cinder.scheduler.driver.Scheduler.host_passes_filters')
    @mock.patch('cinder.db.volume_update')
    def test_migrate_volume_exception_returns_volume_state_maintenance(
            self, _mock_volume_update, _mock_host_passes,
            _mock_volume_get):
        fake_updates = {'status': 'available',
                        'migration_status': 'error'}
        self._test_migrate_volume_exception_returns_volume_state(
            _mock_volume_update, _mock_host_passes, _mock_volume_get,
            'maintenance', fake_updates)

    def _test_migrate_volume_exception_returns_volume_state(
            self, _mock_volume_update, _mock_host_passes,
            _mock_volume_get, status, fake_updates):
        volume = tests_utils.create_volume(self.context,
                                           status=status,
                                           previous_status='available')
        fake_volume_id = volume.id
        topic = 'fake_topic'
        request_spec = {'volume_id': fake_volume_id}
        _mock_host_passes.side_effect = exception.NoValidHost(reason="")
        _mock_volume_get.return_value = volume

        self.manager.migrate_volume_to_host(self.context, topic,
                                            fake_volume_id, 'host', True,
                                            request_spec=request_spec,
                                            filter_properties={})
        _mock_volume_update.assert_called_once_with(self.context,
                                                    fake_volume_id,
                                                    fake_updates)
        _mock_host_passes.assert_called_once_with(self.context, 'host',
                                                  request_spec, {})

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.db.volume_attachment_get_used_by_volume_id')
    def test_retype_volume_exception_returns_volume_state(
            self, _mock_vol_attachment_get, _mock_vol_update):
        # Test NoValidHost exception behavior for retype.
        # Puts the volume in original state and eats the exception.
        volume = tests_utils.create_volume(self.context,
                                           status='retyping',
                                           previous_status='in-use')
        instance_uuid = '12345678-1234-5678-1234-567812345678'
        volume_attach = tests_utils.attach_volume(self.context, volume.id,
                                                  instance_uuid, None,
                                                  '/dev/fake')
        _mock_vol_attachment_get.return_value = [volume_attach]
        topic = 'fake_topic'
        request_spec = {'volume_id': volume.id, 'volume_type': {'id': 3},
                        'migration_policy': 'on-demand'}
        _mock_vol_update.return_value = {'status': 'in-use'}
        _mock_find_retype_host = mock.Mock(
            side_effect=exception.NoValidHost(reason=""))
        orig_retype = self.manager.driver.find_retype_host
        self.manager.driver.find_retype_host = _mock_find_retype_host

        self.manager.retype(self.context, topic, volume.id,
                            request_spec=request_spec,
                            filter_properties={},
                            volume=volume)

        _mock_find_retype_host.assert_called_once_with(self.context,
                                                       request_spec, {},
                                                       'on-demand')
        _mock_vol_update.assert_called_once_with(self.context, volume.id,
                                                 {'status': 'in-use'})
        self.manager.driver.find_retype_host = orig_retype

    def test_create_consistencygroup_exceptions(self):
        with mock.patch.object(filter_scheduler.FilterScheduler,
                               'schedule_create_consistencygroup') as mock_cg:
            original_driver = self.manager.driver
            consistencygroup_obj = \
                fake_consistencygroup.fake_consistencyobject_obj(self.context)
            self.manager.driver = filter_scheduler.FilterScheduler
            LOG = self.mock_object(manager, 'LOG')
            self.stubs.Set(db, 'consistencygroup_update', mock.Mock())

            ex = exception.CinderException('test')
            mock_cg.side_effect = ex
            group_id = fake.consistency_group_id
            self.assertRaises(exception.CinderException,
                              self.manager.create_consistencygroup,
                              self.context,
                              'volume',
                              consistencygroup_obj)
            self.assertTrue(LOG.exception.call_count > 0)
            db.consistencygroup_update.assert_called_once_with(
                self.context, group_id, {'status': (
                    fields.ConsistencyGroupStatus.ERROR)})

            mock_cg.reset_mock()
            LOG.exception.reset_mock()
            db.consistencygroup_update.reset_mock()

            mock_cg.side_effect = exception.NoValidHost(
                reason="No weighed hosts available")
            self.manager.create_consistencygroup(
                self.context, 'volume', consistencygroup_obj)
            self.assertTrue(LOG.error.call_count > 0)
            db.consistencygroup_update.assert_called_once_with(
                self.context, group_id, {'status': (
                    fields.ConsistencyGroupStatus.ERROR)})

            self.manager.driver = original_driver


class SchedulerTestCase(test.TestCase):
    """Test case for base scheduler driver class."""

    # So we can subclass this test and re-use tests if we need.
    driver_cls = driver.Scheduler

    def setUp(self):
        super(SchedulerTestCase, self).setUp()
        self.driver = self.driver_cls()
        self.context = context.RequestContext(fake.user_id, fake.project_id)
        self.topic = 'fake_topic'

    @mock.patch('cinder.scheduler.driver.Scheduler.'
                'update_service_capabilities')
    def test_update_service_capabilities(self, _mock_update_cap):
        service_name = 'fake_service'
        host = 'fake_host'
        capabilities = {'fake_capability': 'fake_value'}
        self.driver.update_service_capabilities(service_name, host,
                                                capabilities)
        _mock_update_cap.assert_called_once_with(service_name, host,
                                                 capabilities)

    @mock.patch('cinder.scheduler.host_manager.HostManager.'
                'has_all_capabilities', return_value=False)
    def test_is_ready(self, _mock_has_caps):
        ready = self.driver.is_ready()
        _mock_has_caps.assert_called_once_with()
        self.assertFalse(ready)


class SchedulerDriverBaseTestCase(SchedulerTestCase):
    """Test schedule driver class.

    Test cases for base scheduler driver class methods
    that will fail if the driver is changed.
    """

    def test_unimplemented_schedule(self):
        fake_args = (1, 2, 3)
        fake_kwargs = {'cat': 'meow'}

        self.assertRaises(NotImplementedError, self.driver.schedule,
                          self.context, self.topic, 'schedule_something',
                          *fake_args, **fake_kwargs)


class SchedulerDriverModuleTestCase(test.TestCase):
    """Test case for scheduler driver module methods."""

    def setUp(self):
        super(SchedulerDriverModuleTestCase, self).setUp()
        self.context = context.RequestContext(fake.user_id, fake.project_id)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    def test_volume_host_update_db(self, _mock_volume_get, _mock_vol_update):
        volume = fake_volume.fake_volume_obj(self.context)
        _mock_volume_get.return_value = volume

        driver.volume_update_db(self.context, volume.id, 'fake_host')
        scheduled_at = volume.scheduled_at.replace(tzinfo=None)
        _mock_vol_update.assert_called_once_with(
            self.context, volume.id, {'host': 'fake_host',
                                      'scheduled_at': scheduled_at})



# Copyright 2013 Canonical Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
""" Tests for Ceph backup service."""

import hashlib
import os
import tempfile
import uuid

import mock
from oslo_concurrency import processutils
from oslo_serialization import jsonutils
import six
from six.moves import range

from cinder.backup import driver
from cinder.backup.drivers import ceph
from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder import test
from cinder.volume.drivers import rbd as rbddriver

# This is used to collect raised exceptions so that tests may check what was
# raised.
# NOTE: this must be initialised in test setUp().
RAISED_EXCEPTIONS = []


class MockException(Exception):

    def __init__(self, *args, **kwargs):
        RAISED_EXCEPTIONS.append(self.__class__)


class MockImageNotFoundException(MockException):
    """Used as mock for rbd.ImageNotFound."""


class MockImageBusyException(MockException):
    """Used as mock for rbd.ImageBusy."""


class MockObjectNotFoundException(MockException):
    """Used as mock for rados.MockObjectNotFoundException."""


def common_mocks(f):
    """Decorator to set mocks common to all tests.

    The point of doing these mocks here is so that we don't accidentally set
    mocks that can't/don't get unset.
    """
    def _common_inner_inner1(inst, *args, **kwargs):
        # NOTE(dosaboy): mock Popen to, by default, raise Exception in order to
        #                ensure that any test ending up in a subprocess fails
        #                if not properly mocked.
        @mock.patch('subprocess.Popen', spec=True)
        # NOTE(dosaboy): mock out eventlet.sleep() so that it does nothing.
        @mock.patch('eventlet.sleep', spec=True)
        @mock.patch('time.time', spec=True)
        # NOTE(dosaboy): set spec to empty object so that hasattr calls return
        #                False by default.
        @mock.patch('cinder.backup.drivers.ceph.rbd')
        @mock.patch('cinder.backup.drivers.ceph.rados')
        def _common_inner_inner2(mock_rados, mock_rbd, mock_time, mock_sleep,
                                 mock_popen):
            mock_time.side_effect = inst.time_inc
            mock_popen.side_effect = Exception

            inst.mock_rados = mock_rados
            inst.mock_rbd = mock_rbd
            inst.mock_rbd.ImageBusy = MockImageBusyException
            inst.mock_rbd.ImageNotFound = MockImageNotFoundException

            inst.service.rbd = inst.mock_rbd
            inst.service.rados = inst.mock_rados
            return f(inst, *args, **kwargs)

        return _common_inner_inner2()

    return _common_inner_inner1


class BackupCephTestCase(test.TestCase):
    """Test case for ceph backup driver."""

    def _create_volume_db_entry(self, id, size):
        vol = {'id': id, 'size': size, 'status': 'available'}
        return db.volume_create(self.ctxt, vol)['id']

    def _create_backup_db_entry(self, backupid, volid, size,
                                userid=str(uuid.uuid4()),
                                projectid=str(uuid.uuid4())):
        backup = {'id': backupid, 'size': size, 'volume_id': volid,
                  'user_id': userid, 'project_id': projectid}
        return db.backup_create(self.ctxt, backup)['id']

    def time_inc(self):
        self.counter += 1
        return self.counter

    def _get_wrapped_rbd_io(self, rbd_image):
        rbd_meta = rbddriver.RBDImageMetadata(rbd_image, 'pool_foo',
                                              'user_foo', 'conf_foo')
        return rbddriver.RBDImageIOWrapper(rbd_meta)

    def _setup_mock_popen(self, mock_popen, retval=None, p1hook=None,
                          p2hook=None):

        class MockPopen(object):
            hooks = [p2hook, p1hook]

            def __init__(mock_inst, cmd, *args, **kwargs):
                self.callstack.append('popen_init')
                mock_inst.stdout = mock.Mock()
                mock_inst.stdout.close = mock.Mock()
                mock_inst.stdout.close.side_effect = \
                    lambda *args: self.callstack.append('stdout_close')
                mock_inst.returncode = 0
                hook = mock_inst.__class__.hooks.pop()
                if hook is not None:
                    hook()

            def communicate(mock_inst):
                self.callstack.append('communicate')
                return retval

        mock_popen.side_effect = MockPopen

    def setUp(self):
        global RAISED_EXCEPTIONS
        RAISED_EXCEPTIONS = []
        super(BackupCephTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

        # Create volume.
        self.volume_size = 1
        self.volume_id = str(uuid.uuid4())
        self._create_volume_db_entry(self.volume_id, self.volume_size)
        self.volume = db.volume_get(self.ctxt, self.volume_id)

        # Create backup of volume.
        self.backup_id = str(uuid.uuid4())
        self._create_backup_db_entry(self.backup_id, self.volume_id,
                                     self.volume_size)
        self.backup = objects.Backup.get_by_id(self.ctxt, self.backup_id)

        # Create alternate volume.
        self.alt_volume_id = str(uuid.uuid4())
        self._create_volume_db_entry(self.alt_volume_id, self.volume_size)
        self.alt_volume = db.volume_get(self.ctxt, self.alt_volume_id)

        self.chunk_size = 1024
        self.num_chunks = 128
        self.data_length = self.num_chunks * self.chunk_size
        self.checksum = hashlib.sha256()

        # Create a file with some data in it.
        self.volume_file = tempfile.NamedTemporaryFile()
        self.addCleanup(self.volume_file.close)
        for _i in range(0, self.num_chunks):
            data = os.urandom(self.chunk_size)
            self.checksum.update(data)
            self.volume_file.write(data)

        self.volume_file.seek(0)

        # Always trigger an exception if a command is executed since it should
        # always be dealt with gracefully. At time of writing on rbd
        # export/import-diff is executed and if they fail we expect to find
        # alternative means of backing up.
        mock_exec = mock.Mock()
        mock_exec.side_effect = processutils.ProcessExecutionError

        self.service = ceph.CephBackupDriver(self.ctxt, execute=mock_exec)

        # Ensure that time.time() always returns more than the last time it was
        # called to avoid div by zero errors.
        self.counter = float(0)

        self.callstack = []

    @common_mocks
    def test_get_rbd_support(self):
        del self.service.rbd.RBD_FEATURE_LAYERING
        del self.service.rbd.RBD_FEATURE_STRIPINGV2
        self.assertFalse(hasattr(self.service.rbd, 'RBD_FEATURE_LAYERING'))
        self.assertFalse(hasattr(self.service.rbd, 'RBD_FEATURE_STRIPINGV2'))

        oldformat, features = self.service._get_rbd_support()
        self.assertTrue(oldformat)
        self.assertEqual(0, features)

        self.service.rbd.RBD_FEATURE_LAYERING = 1

        oldformat, features = self.service._get_rbd_support()
        self.assertFalse(oldformat)
        self.assertEqual(1, features)

        self.service.rbd.RBD_FEATURE_STRIPINGV2 = 2

        oldformat, features = self.service._get_rbd_support()
        self.assertFalse(oldformat)
        self.assertEqual(1 | 2, features)

    @common_mocks
    def test_get_most_recent_snap(self):
        last = 'backup.%s.snap.9824923.1212' % (uuid.uuid4())

        image = self.mock_rbd.Image.return_value
        image.list_snaps.return_value = \
            [{'name': 'backup.%s.snap.6423868.2342' % (uuid.uuid4())},
             {'name': 'backup.%s.snap.1321319.3235' % (uuid.uuid4())},
             {'name': last},
             {'name': 'backup.%s.snap.3824923.1412' % (uuid.uuid4())}]

        snap = self.service._get_most_recent_snap(image)
        self.assertEqual(last, snap)

    @common_mocks
    def test_get_backup_snap_name(self):
        snap_name = 'backup.%s.snap.3824923.1412' % (uuid.uuid4())

        def get_backup_snaps(inst, *args):
            return [{'name': 'backup.%s.snap.6423868.2342' % (uuid.uuid4()),
                     'backup_id': str(uuid.uuid4())},
                    {'name': snap_name,
                     'backup_id': self.backup_id}]

        with mock.patch.object(self.service, 'get_backup_snaps'):
            name = self.service._get_backup_snap_name(self.service.rbd.Image(),
                                                      'base_foo',
                                                      self.backup_id)
            self.assertIsNone(name)

        with mock.patch.object(self.service, 'get_backup_snaps') as \
                mock_get_backup_snaps:
            mock_get_backup_snaps.side_effect = get_backup_snaps
            name = self.service._get_backup_snap_name(self.service.rbd.Image(),
                                                      'base_foo',
                                                      self.backup_id)
            self.assertEqual(snap_name, name)
            self.assertTrue(mock_get_backup_snaps.called)

    @common_mocks
    def test_get_backup_snaps(self):
        image = self.mock_rbd.Image.return_value
        image.list_snaps.return_value = [
            {'name': 'backup.%s.snap.6423868.2342' % (uuid.uuid4())},
            {'name': 'backup.%s.wambam.6423868.2342' % (uuid.uuid4())},
            {'name': 'backup.%s.snap.1321319.3235' % (uuid.uuid4())},
            {'name': 'bbbackup.%s.snap.1321319.3235' % (uuid.uuid4())},
            {'name': 'backup.%s.snap.3824923.1412' % (uuid.uuid4())}]
        snaps = self.service.get_backup_snaps(image)
        self.assertEqual(3, len(snaps))

    @common_mocks
    def test_transfer_data_from_rbd_to_file(self):
        def fake_read(offset, length):
            self.volume_file.seek(offset)
            return self.volume_file.read(length)

        self.mock_rbd.Image.return_value.read.side_effect = fake_read
        self.mock_rbd.Image.return_value.size.return_value = self.data_length

        with tempfile.NamedTemporaryFile() as test_file:
            self.volume_file.seek(0)

            rbd_io = self._get_wrapped_rbd_io(self.service.rbd.Image())
            self.service._transfer_data(rbd_io, 'src_foo', test_file,
                                        'dest_foo', self.data_length)

            checksum = hashlib.sha256()
            test_file.seek(0)
            for _c in range(0, self.num_chunks):
                checksum.update(test_file.read(self.chunk_size))

            # Ensure the files are equal
            self.assertEqual(checksum.digest(), self.checksum.digest())

    @common_mocks
    def test_transfer_data_from_rbd_to_rbd(self):
        def fake_read(offset, length):
            self.volume_file.seek(offset)
            return self.volume_file.read(length)

        def mock_write_data(data, offset):
            checksum.update(data)
            test_file.write(data)

        rbd1 = mock.Mock()
        rbd1.read.side_effect = fake_read
        rbd1.size.return_value = os.fstat(self.volume_file.fileno()).st_size

        rbd2 = mock.Mock()
        rbd2.write.side_effect = mock_write_data

        with tempfile.NamedTemporaryFile() as test_file:
            self.volume_file.seek(0)
            checksum = hashlib.sha256()

            src_rbd_io = self._get_wrapped_rbd_io(rbd1)
            dest_rbd_io = self._get_wrapped_rbd_io(rbd2)
            self.service._transfer_data(src_rbd_io, 'src_foo', dest_rbd_io,
                                        'dest_foo', self.data_length)

            # Ensure the files are equal
            self.assertEqual(checksum.digest(), self.checksum.digest())

    @common_mocks
    def test_transfer_data_from_file_to_rbd(self):

        def mock_write_data(data, offset):
            checksum.update(data)
            test_file.write(data)

        self.mock_rbd.Image.return_value.write.side_effect = mock_write_data

        with tempfile.NamedTemporaryFile() as test_file:
            self.volume_file.seek(0)
            checksum = hashlib.sha256()

            rbd_io = self._get_wrapped_rbd_io(self.service.rbd.Image())
            self.service._transfer_data(self.volume_file, 'src_foo',
                                        rbd_io, 'dest_foo', self.data_length)

            # Ensure the files are equal
            self.assertEqual(checksum.digest(), self.checksum.digest())

    @common_mocks
    def test_transfer_data_from_file_to_file(self):
        with tempfile.NamedTemporaryFile() as test_file:
            self.volume_file.seek(0)
            checksum = hashlib.sha256()

            self.service._transfer_data(self.volume_file, 'src_foo', test_file,
                                        'dest_foo', self.data_length)

            checksum = hashlib.sha256()
            test_file.seek(0)
            for _c in range(0, self.num_chunks):
                checksum.update(test_file.read(self.chunk_size))

            # Ensure the files are equal
            self.assertEqual(checksum.digest(), self.checksum.digest())

    @common_mocks
    def test_backup_volume_from_file(self):
        checksum = hashlib.sha256()

        def mock_write_data(data, offset):
            checksum.update(data)
            test_file.write(data)

        self.service.rbd.Image.return_value.write.side_effect = mock_write_data

        with mock.patch.object(self.service, '_backup_metadata'):
            with mock.patch.object(self.service, '_discard_bytes'):
                with tempfile.NamedTemporaryFile() as test_file:
                    self.service.backup(self.backup, self.volume_file)

                    # Ensure the files are equal
                    self.assertEqual(checksum.digest(), self.checksum.digest())

        self.assertTrue(self.service.rbd.Image.return_value.write.called)

    @common_mocks
    def test_get_backup_base_name(self):
        name = self.service._get_backup_base_name(self.volume_id,
                                                  diff_format=True)
        self.assertEqual("volume-%s.backup.base" % (self.volume_id), name)

        self.assertRaises(exception.InvalidParameterValue,
                          self.service._get_backup_base_name,
                          self.volume_id)

        name = self.service._get_backup_base_name(self.volume_id, '1234')
        self.assertEqual("volume-%s.backup.%s" % (self.volume_id, '1234'),
                         name)

    @common_mocks
    @mock.patch('fcntl.fcntl', spec=True)
    @mock.patch('subprocess.Popen', spec=True)
    def test_backup_volume_from_rbd(self, mock_popen, mock_fnctl):
        backup_name = self.service._get_backup_base_name(self.backup_id,
                                                         diff_format=True)

        def mock_write_data():
            self.volume_file.seek(0)
            data = self.volume_file.read(self.data_length)
            self.callstack.append('write')
            checksum.update(data)
            test_file.write(data)

        def mock_read_data():
            self.callstack.append('read')
            return self.volume_file.read(self.data_length)

        self._setup_mock_popen(mock_popen,
                               ['out', 'err'],
                               p1hook=mock_read_data,
                               p2hook=mock_write_data)

        self.mock_rbd.RBD.list = mock.Mock()
        self.mock_rbd.RBD.list.return_value = [backup_name]

        with mock.patch.object(self.service, '_backup_metadata'):
            with mock.patch.object(self.service, 'get_backup_snaps') as \
                    mock_get_backup_snaps:
                with mock.patch.object(self.service, '_full_backup') as \
                        mock_full_backup:
                    with mock.patch.object(self.service,
                                           '_try_delete_base_image'):
                        with tempfile.NamedTemporaryFile() as test_file:
                            checksum = hashlib.sha256()
                            image = self.service.rbd.Image()
                            meta = rbddriver.RBDImageMetadata(image,
                                                              'pool_foo',
                                                              'user_foo',
                                                              'conf_foo')
                            rbdio = rbddriver.RBDImageIOWrapper(meta)
                            self.service.backup(self.backup, rbdio)

                            self.assertEqual(['popen_init',
                                              'read',
                                              'popen_init',
                                              'write',
                                              'stdout_close',
                                              'communicate'], self.callstack)

                            self.assertFalse(mock_full_backup.called)
                            self.assertTrue(mock_get_backup_snaps.called)

                            # Ensure the files are equal
                            self.assertEqual(checksum.digest(),
                                             self.checksum.digest())

    @common_mocks
    @mock.patch('fcntl.fcntl', spec=True)
    @mock.patch('subprocess.Popen', spec=True)
    def test_backup_volume_from_rbd_fail(self, mock_popen, mock_fnctl):
        """Test of when an exception occurs in an exception handler.

        In _backup_rbd(), after an exception.BackupRBDOperationFailed
        occurs in self._rbd_diff_transfer(), we want to check the
        process when the second exception occurs in
        self._try_delete_base_image().
        """
        backup_name = self.service._get_backup_base_name(self.backup_id,
                                                         diff_format=True)

        def mock_write_data():
            self.volume_file.seek(0)
            data = self.volume_file.read(self.data_length)
            self.callstack.append('write')
            checksum.update(data)
            test_file.write(data)

        def mock_read_data():
            self.callstack.append('read')
            return self.volume_file.read(self.data_length)

        self._setup_mock_popen(mock_popen,
                               ['out', 'err'],
                               p1hook=mock_read_data,
                               p2hook=mock_write_data)

        self.mock_rbd.RBD.list = mock.Mock()
        self.mock_rbd.RBD.list.return_value = [backup_name]

        with mock.patch.object(self.service, 'get_backup_snaps'), \
                mock.patch.object(self.service, '_rbd_diff_transfer') as \
                mock_rbd_diff_transfer:
            def mock_rbd_diff_transfer_side_effect(src_name, src_pool,
                                                   dest_name, dest_pool,
                                                   src_user, src_conf,
                                                   dest_user, dest_conf,
                                                   src_snap, from_snap):
                raise exception.BackupRBDOperationFailed(_('mock'))

            # Raise a pseudo exception.BackupRBDOperationFailed.
            mock_rbd_diff_transfer.side_effect \
                = mock_rbd_diff_transfer_side_effect

            with mock.patch.object(self.service, '_full_backup'), \
                    mock.patch.object(self.service,
                                      '_try_delete_base_image') as \
                    mock_try_delete_base_image:
                def mock_try_delete_base_image_side_effect(backup_id,
                                                           volume_id,
                                                           base_name):
                    raise self.service.rbd.ImageNotFound(_('mock'))

                # Raise a pesudo exception rbd.ImageNotFound.
                mock_try_delete_base_image.side_effect \
                    = mock_try_delete_base_image_side_effect
                with mock.patch.object(self.service, '_backup_metadata'):
                    with tempfile.NamedTemporaryFile() as test_file:
                        checksum = hashlib.sha256()
                        image = self.service.rbd.Image()
                        meta = rbddriver.RBDImageMetadata(image,
                                                          'pool_foo',
                                                          'user_foo',
                                                          'conf_foo')
                        rbdio = rbddriver.RBDImageIOWrapper(meta)

                        # We expect that the second exception is
                        # notified.
                        self.assertRaises(
                            self.service.rbd.ImageNotFound,
                            self.service.backup,
                            self.backup, rbdio)

    @common_mocks
    @mock.patch('fcntl.fcntl', spec=True)
    @mock.patch('subprocess.Popen', spec=True)
    def test_backup_volume_from_rbd_fail2(self, mock_popen, mock_fnctl):
        """Test of when an exception occurs in an exception handler.

        In backup(), after an exception.BackupOperationError occurs in
        self._backup_metadata(), we want to check the process when the
        second exception occurs in self.delete().
        """
        backup_name = self.service._get_backup_base_name(self.backup_id,
                                                         diff_format=True)

        def mock_write_data():
            self.volume_file.seek(0)
            data = self.volume_file.read(self.data_length)
            self.callstack.append('write')
            checksum.update(data)
            test_file.write(data)

        def mock_read_data():
            self.callstack.append('read')
            return self.volume_file.read(self.data_length)

        self._setup_mock_popen(mock_popen,
                               ['out', 'err'],
                               p1hook=mock_read_data,
                               p2hook=mock_write_data)

        self.mock_rbd.RBD.list = mock.Mock()
        self.mock_rbd.RBD.list.return_value = [backup_name]

        with mock.patch.object(self.service, 'get_backup_snaps'), \
                mock.patch.object(self.service, '_rbd_diff_transfer'), \
                mock.patch.object(self.service, '_full_backup'), \
                mock.patch.object(self.service, '_backup_metadata') as \
                mock_backup_metadata:

            def mock_backup_metadata_side_effect(backup):
                raise exception.BackupOperationError(_('mock'))

            # Raise a pseudo exception.BackupOperationError.
            mock_backup_metadata.side_effect = mock_backup_metadata_side_effect
            with mock.patch.object(self.service, 'delete') as mock_delete:
                def mock_delete_side_effect(backup):
                    raise self.service.rbd.ImageBusy()

                # Raise a pseudo exception rbd.ImageBusy.
                mock_delete.side_effect = mock_delete_side_effect
                with tempfile.NamedTemporaryFile() as test_file:
                    checksum = hashlib.sha256()
                    image = self.service.rbd.Image()
                    meta = rbddriver.RBDImageMetadata(image,
                                                      'pool_foo',
                                                      'user_foo',
                                                      'conf_foo')
                    rbdio = rbddriver.RBDImageIOWrapper(meta)

                    # We expect that the second exception is
                    # notified.
                    self.assertRaises(
                        self.service.rbd.ImageBusy,
                        self.service.backup,
                        self.backup, rbdio)

    @common_mocks
    def test_backup_vol_length_0(self):
        volume_id = str(uuid.uuid4())
        self._create_volume_db_entry(volume_id, 0)

        backup_id = str(uuid.uuid4())
        self._create_backup_db_entry(backup_id, volume_id, 1)
        backup = objects.Backup.get_by_id(self.ctxt, backup_id)

        self.assertRaises(exception.InvalidParameterValue, self.service.backup,
                          backup, self.volume_file)

    @common_mocks
    def test_restore(self):
        backup_name = self.service._get_backup_base_name(self.backup_id,
                                                         diff_format=True)

        self.mock_rbd.RBD.return_value.list.return_value = [backup_name]

        def mock_read_data(offset, length):
            return self.volume_file.read(self.data_length)

        self.mock_rbd.Image.return_value.read.side_effect = mock_read_data

        self.mock_rbd.Image.return_value.size.return_value = \
            self.chunk_size * self.num_chunks

        with mock.patch.object(self.service, '_restore_metadata') as \
                mock_restore_metadata:
            with mock.patch.object(self.service, '_discard_bytes') as \
                    mock_discard_bytes:
                with tempfile.NamedTemporaryFile() as test_file:
                    self.volume_file.seek(0)

                    self.service.restore(self.backup, self.volume_id,
                                         test_file)

                    checksum = hashlib.sha256()
                    test_file.seek(0)
                    for _c in range(0, self.num_chunks):
                        checksum.update(test_file.read(self.chunk_size))

                    # Ensure the files are equal
                    self.assertEqual(checksum.digest(), self.checksum.digest())

                    self.assertTrue(mock_restore_metadata.called)
                    self.assertTrue(mock_discard_bytes.called)
                    self.assertTrue(mock_discard_bytes.called)

        self.assertTrue(self.service.rbd.Image.return_value.read.called)

    @common_mocks
    def test_discard_bytes(self):
        # Lower the chunksize to a memory managable number
        self.service.chunk_size = 1024
        image = self.mock_rbd.Image.return_value
        wrapped_rbd = self._get_wrapped_rbd_io(image)

        self.service._discard_bytes(wrapped_rbd, 0, 0)
        self.assertEqual(0, image.discard.call_count)

        self.service._discard_bytes(wrapped_rbd, 0, 1234)
        self.assertEqual(1, image.discard.call_count)
        image.reset_mock()

        # Test discard with no remainder
        with mock.patch.object(self.service, '_file_is_rbd') as \
                mock_file_is_rbd:
            mock_file_is_rbd.return_value = False

            self.service._discard_bytes(wrapped_rbd, 0,
                                        self.service.chunk_size * 2)

            self.assertEqual(2, image.write.call_count)
            self.assertEqual(2, image.flush.call_count)
            self.assertFalse(image.discard.called)

        image.reset_mock()

        # Now test with a remainder.
        with mock.patch.object(self.service, '_file_is_rbd') as \
                mock_file_is_rbd:
            mock_file_is_rbd.return_value = False

            self.service._discard_bytes(wrapped_rbd, 0,
                                        (self.service.chunk_size * 2) + 1)

            self.assertEqual(3, image.write.call_count)
            self.assertEqual(3, image.flush.call_count)
            self.assertFalse(image.discard.called)

    @common_mocks
    def test_delete_backup_snapshot(self):
        snap_name = 'backup.%s.snap.3824923.1412' % (uuid.uuid4())
        base_name = self.service._get_backup_base_name(self.volume_id,
                                                       diff_format=True)
        self.mock_rbd.RBD.remove_snap = mock.Mock()

        with mock.patch.object(self.service, '_get_backup_snap_name') as \
                mock_get_backup_snap_name:
            mock_get_backup_snap_name.return_value = snap_name
            with mock.patch.object(self.service, 'get_backup_snaps') as \
                    mock_get_backup_snaps:
                mock_get_backup_snaps.return_value = None
                rem = self.service._delete_backup_snapshot(self.mock_rados,
                                                           base_name,
                                                           self.backup_id)

                self.assertTrue(mock_get_backup_snap_name.called)
                self.assertTrue(mock_get_backup_snaps.called)
                self.assertEqual((snap_name, 0), rem)

    @common_mocks
    @mock.patch('cinder.backup.drivers.ceph.VolumeMetadataBackup', spec=True)
    def test_try_delete_base_image_diff_format(self, mock_meta_backup):
        backup_name = self.service._get_backup_base_name(self.volume_id,
                                                         diff_format=True)

        self.mock_rbd.RBD.return_value.list.return_value = [backup_name]

        with mock.patch.object(self.service, '_delete_backup_snapshot') as \
                mock_del_backup_snap:
            snap_name = self.service._get_new_snap_name(self.backup_id)
            mock_del_backup_snap.return_value = (snap_name, 0)

            self.service.delete(self.backup)
            self.assertTrue(mock_del_backup_snap.called)

        self.assertTrue(self.mock_rbd.RBD.return_value.list.called)
        self.assertTrue(self.mock_rbd.RBD.return_value.remove.called)

    @common_mocks
    @mock.patch('cinder.backup.drivers.ceph.VolumeMetadataBackup', spec=True)
    def test_try_delete_base_image(self, mock_meta_backup):
        backup_name = self.service._get_backup_base_name(self.volume_id,
                                                         self.backup_id)

        self.mock_rbd.RBD.return_value.list.return_value = [backup_name]

        with mock.patch.object(self.service, 'get_backup_snaps'):
            self.service.delete(self.backup)
            self.assertTrue(self.mock_rbd.RBD.return_value.remove.called)

    @common_mocks
    def test_try_delete_base_image_busy(self):
        """This should induce retries then raise rbd.ImageBusy."""
        backup_name = self.service._get_backup_base_name(self.volume_id,
                                                         self.backup_id)

        rbd = self.mock_rbd.RBD.return_value
        rbd.list.return_value = [backup_name]
        rbd.remove.side_effect = self.mock_rbd.ImageBusy

        with mock.patch.object(self.service, 'get_backup_snaps') as \
                mock_get_backup_snaps:
            self.assertRaises(self.mock_rbd.ImageBusy,
                              self.service._try_delete_base_image,
                              self.backup['id'], self.backup['volume_id'])
            self.assertTrue(mock_get_backup_snaps.called)

        self.assertTrue(rbd.list.called)
        self.assertTrue(rbd.remove.called)
        self.assertTrue(MockImageBusyException in RAISED_EXCEPTIONS)

    @common_mocks
    @mock.patch('cinder.backup.drivers.ceph.VolumeMetadataBackup', spec=True)
    def test_delete(self, mock_meta_backup):
        with mock.patch.object(self.service, '_try_delete_base_image'):
            self.service.delete(self.backup)
            self.assertEqual([], RAISED_EXCEPTIONS)

    @common_mocks
    @mock.patch('cinder.backup.drivers.ceph.VolumeMetadataBackup', spec=True)
    def test_delete_image_not_found(self, mock_meta_backup):
        with mock.patch.object(self.service, '_try_delete_base_image') as \
                mock_del_base:
            mock_del_base.side_effect = self.mock_rbd.ImageNotFound
            # ImageNotFound exception is caught so that db entry can be cleared
            self.service.delete(self.backup)
            self.assertEqual([MockImageNotFoundException], RAISED_EXCEPTIONS)

    @common_mocks
    def test_diff_restore_allowed_with_image_not_exists(self):
        """Test diff restore not allowed when backup not diff-format."""
        not_allowed = (False, None)
        backup_base = 'backup.base'
        rbd_io = self._get_wrapped_rbd_io(self.service.rbd.Image())
        args_vols_different = [backup_base, self.backup, self.alt_volume,
                               rbd_io, self.mock_rados]

        with mock.patch.object(self.service, '_rbd_image_exists') as \
                mock_rbd_image_exists:
            mock_rbd_image_exists.return_value = (False, backup_base)

            resp = self.service._diff_restore_allowed(*args_vols_different)

            self.assertEqual(not_allowed, resp)
            mock_rbd_image_exists.assert_called_once_with(
                backup_base,
                self.backup['volume_id'],
                self.mock_rados)

    @common_mocks
    def test_diff_restore_allowed_with_no_restore_point(self):
        """Test diff restore not allowed when no restore point found.

        Detail conditions:
          1. backup base is diff-format
          2. restore point does not exist
        """
        not_allowed = (False, None)
        backup_base = 'backup.base'
        rbd_io = self._get_wrapped_rbd_io(self.service.rbd.Image())
        args_vols_different = [backup_base, self.backup, self.alt_volume,
                               rbd_io, self.mock_rados]

        with mock.patch.object(self.service, '_rbd_image_exists') as \
                mock_rbd_image_exists:
            mock_rbd_image_exists.return_value = (True, backup_base)
            with mock.patch.object(self.service, '_get_restore_point') as \
                    mock_get_restore_point:
                mock_get_restore_point.return_value = None

                args = args_vols_different
                resp = self.service._diff_restore_allowed(*args)

                self.assertEqual(not_allowed, resp)
                self.assertTrue(mock_rbd_image_exists.called)
                mock_get_restore_point.assert_called_once_with(
                    backup_base,
                    self.backup['id'])

    @common_mocks
    def test_diff_restore_allowed_with_not_rbd(self):
        """Test diff restore not allowed when destination volume is not rbd.

        Detail conditions:
          1. backup base is diff-format
          2. restore point exists
          3. destination volume is not an rbd.
        """
        backup_base = 'backup.base'
        restore_point = 'backup.snap.1'
        rbd_io = self._get_wrapped_rbd_io(self.service.rbd.Image())
        args_vols_different = [backup_base, self.backup, self.alt_volume,
                               rbd_io, self.mock_rados]

        with mock.patch.object(self.service, '_rbd_image_exists') as \
                mock_rbd_image_exists:
            mock_rbd_image_exists.return_value = (True, backup_base)
            with mock.patch.object(self.service, '_get_restore_point') as \
                    mock_get_restore_point:
                mock_get_restore_point.return_value = restore_point
                with mock.patch.object(self.service, '_file_is_rbd') as \
                        mock_file_is_rbd:
                    mock_file_is_rbd.return_value = False

                    args = args_vols_different
                    resp = self.service._diff_restore_allowed(*args)

                    self.assertEqual((False, restore_point), resp)
                    self.assertTrue(mock_rbd_image_exists.called)
                    self.assertTrue(mock_get_restore_point.called)
                    mock_file_is_rbd.assert_called_once_with(
                        rbd_io)

    @common_mocks
    def test_diff_restore_allowed_with_same_volume(self):
        """Test diff restore not allowed when volumes are same.

        Detail conditions:
          1. backup base is diff-format
          2. restore point exists
          3. destination volume is an rbd
          4. source and destination volumes are the same
        """
        backup_base = 'backup.base'
        restore_point = 'backup.snap.1'
        rbd_io = self._get_wrapped_rbd_io(self.service.rbd.Image())
        args_vols_same = [backup_base, self.backup, self.volume, rbd_io,
                          self.mock_rados]

        with mock.patch.object(self.service, '_rbd_image_exists') as \
                mock_rbd_image_exists:
            mock_rbd_image_exists.return_value = (True, backup_base)
            with mock.patch.object(self.service, '_get_restore_point') as \
                    mock_get_restore_point:
                mock_get_restore_point.return_value = restore_point
                with mock.patch.object(self.service, '_file_is_rbd') as \
                        mock_file_is_rbd:
                    mock_file_is_rbd.return_value = True

                    resp = self.service._diff_restore_allowed(*args_vols_same)

                    self.assertEqual((False, restore_point), resp)
                    self.assertTrue(mock_rbd_image_exists.called)
                    self.assertTrue(mock_get_restore_point.called)
                    self.assertTrue(mock_file_is_rbd.called)

    @common_mocks
    def test_diff_restore_allowed_with_has_extents(self):
        """Test diff restore not allowed when destination volume has data.

        Detail conditions:
          1. backup base is diff-format
          2. restore point exists
          3. destination volume is an rbd
          4. source and destination volumes are different
          5. destination volume has data on it - full copy is mandated
        """
        backup_base = 'backup.base'
        restore_point = 'backup.snap.1'
        rbd_io = self._get_wrapped_rbd_io(self.service.rbd.Image())
        args_vols_different = [backup_base, self.backup, self.alt_volume,
                               rbd_io, self.mock_rados]

        with mock.patch.object(self.service, '_rbd_image_exists') as \
                mock_rbd_image_exists:
            mock_rbd_image_exists.return_value = (True, backup_base)
            with mock.patch.object(self.service, '_get_restore_point') as \
                    mock_get_restore_point:
                mock_get_restore_point.return_value = restore_point
                with mock.patch.object(self.service, '_file_is_rbd') as \
                        mock_file_is_rbd:
                    mock_file_is_rbd.return_value = True
                    with mock.patch.object(self.service, '_rbd_has_extents') \
                            as mock_rbd_has_extents:
                        mock_rbd_has_extents.return_value = True

                        args = args_vols_different
                        resp = self.service._diff_restore_allowed(*args)

                        self.assertEqual((False, restore_point), resp)
                        self.assertTrue(mock_rbd_image_exists.called)
                        self.assertTrue(mock_get_restore_point.called)
                        self.assertTrue(mock_file_is_rbd.called)
                        mock_rbd_has_extents.assert_called_once_with(
                            rbd_io.rbd_image)

    @common_mocks
    def test_diff_restore_allowed_with_no_extents(self):
        """Test diff restore allowed when no data in destination volume.

        Detail conditions:
          1. backup base is diff-format
          2. restore point exists
          3. destination volume is an rbd
          4. source and destination volumes are different
          5. destination volume no data on it
        """
        backup_base = 'backup.base'
        restore_point = 'backup.snap.1'
        rbd_io = self._get_wrapped_rbd_io(self.service.rbd.Image())
        args_vols_different = [backup_base, self.backup, self.alt_volume,
                               rbd_io, self.mock_rados]

        with mock.patch.object(self.service, '_rbd_image_exists') as \
                mock_rbd_image_exists:
            mock_rbd_image_exists.return_value = (True, backup_base)
            with mock.patch.object(self.service, '_get_restore_point') as \
                    mock_get_restore_point:
                mock_get_restore_point.return_value = restore_point
                with mock.patch.object(self.service, '_file_is_rbd') as \
                        mock_file_is_rbd:
                    mock_file_is_rbd.return_value = True
                    with mock.patch.object(self.service, '_rbd_has_extents') \
                            as mock_rbd_has_extents:
                        mock_rbd_has_extents.return_value = False

                        args = args_vols_different
                        resp = self.service._diff_restore_allowed(*args)

                        self.assertEqual((True, restore_point), resp)
                        self.assertTrue(mock_rbd_image_exists.called)
                        self.assertTrue(mock_get_restore_point.called)
                        self.assertTrue(mock_file_is_rbd.called)
                        self.assertTrue(mock_rbd_has_extents.called)

    @common_mocks
    @mock.patch('fcntl.fcntl', spec=True)
    @mock.patch('subprocess.Popen', spec=True)
    def test_piped_execute(self, mock_popen, mock_fcntl):
        mock_fcntl.return_value = 0
        self._setup_mock_popen(mock_popen, ['out', 'err'])
        self.service._piped_execute(['foo'], ['bar'])
        self.assertEqual(['popen_init', 'popen_init',
                          'stdout_close', 'communicate'], self.callstack)

    @common_mocks
    def test_restore_metdata(self):
        version = 2

        def mock_read(*args):
            base_tag = driver.BackupMetadataAPI.TYPE_TAG_VOL_BASE_META
            glance_tag = driver.BackupMetadataAPI.TYPE_TAG_VOL_GLANCE_META
            return jsonutils.dumps({base_tag: {'image_name': 'image.base'},
                                    glance_tag: {'image_name': 'image.glance'},
                                    'version': version})

        self.mock_rados.Object.return_value.read.side_effect = mock_read

        self.service._restore_metadata(self.backup, self.volume_id)

        self.assertTrue(self.mock_rados.Object.return_value.stat.called)
        self.assertTrue(self.mock_rados.Object.return_value.read.called)

        version = 3
        try:
            self.service._restore_metadata(self.backup, self.volume_id)
        except exception.BackupOperationError as exc:
            msg = _("Metadata restore failed due to incompatible version")
            self.assertEqual(msg, six.text_type(exc))
        else:
            # Force a test failure
            self.assertFalse(True)

    @common_mocks
    @mock.patch('cinder.backup.drivers.ceph.VolumeMetadataBackup', spec=True)
    def test_backup_metadata_already_exists(self, mock_meta_backup):

        def mock_set(json_meta):
            msg = (_("Metadata backup object '%s' already exists") %
                   ("backup.%s.meta" % (self.backup_id)))
            raise exception.VolumeMetadataBackupExists(msg)

        mock_meta_backup.return_value.set = mock.Mock()
        mock_meta_backup.return_value.set.side_effect = mock_set

        with mock.patch.object(self.service, 'get_metadata') as \
                mock_get_metadata:
            mock_get_metadata.return_value = "some.json.metadata"
            try:
                self.service._backup_metadata(self.backup)
            except exception.BackupOperationError as e:
                msg = (_("Failed to backup volume metadata - Metadata backup "
                         "object 'backup.%s.meta' already exists") %
                       (self.backup_id))
                self.assertEqual(msg, six.text_type(e))
            else:
                # Make the test fail
                self.assertFalse(True)

        self.assertFalse(mock_meta_backup.set.called)

    @common_mocks
    def test_backup_metata_error(self):
        """Ensure that delete() is called if the metadata backup fails.

        Also ensure that the exception is propagated to the caller.
        """
        with mock.patch.object(self.service, '_backup_metadata') as \
                mock_backup_metadata:
            mock_backup_metadata.side_effect = exception.BackupOperationError
            with mock.patch.object(self.service, '_get_volume_size_gb'):
                with mock.patch.object(self.service, '_file_is_rbd',
                                       return_value=False):
                    with mock.patch.object(self.service, '_full_backup'):
                        with mock.patch.object(self.service, 'delete') as \
                                mock_delete:
                            self.assertRaises(exception.BackupOperationError,
                                              self.service.backup, self.backup,
                                              mock.Mock(),
                                              backup_metadata=True)
                            self.assertTrue(mock_delete.called)

    @common_mocks
    def test_restore_invalid_metadata_version(self):

        def mock_read(*args):
            base_tag = driver.BackupMetadataAPI.TYPE_TAG_VOL_BASE_META
            glance_tag = driver.BackupMetadataAPI.TYPE_TAG_VOL_GLANCE_META
            return jsonutils.dumps({base_tag: {'image_name': 'image.base'},
                                    glance_tag: {'image_name': 'image.glance'},
                                    'version': 3})

        self.mock_rados.Object.return_value.read.side_effect = mock_read
        with mock.patch.object(ceph.VolumeMetadataBackup, '_exists') as \
                mock_exists:
            mock_exists.return_value = True

            self.assertRaises(exception.BackupOperationError,
                              self.service._restore_metadata,
                              self.backup, self.volume_id)

            self.assertTrue(mock_exists.called)

        self.assertTrue(self.mock_rados.Object.return_value.read.called)


def common_meta_backup_mocks(f):
    """Decorator to set mocks common to all metadata backup tests.

    The point of doing these mocks here is so that we don't accidentally set
    mocks that can't/don't get unset.
    """
    def _common_inner_inner1(inst, *args, **kwargs):
        @mock.patch('cinder.backup.drivers.ceph.rbd')
        @mock.patch('cinder.backup.drivers.ceph.rados')
        def _common_inner_inner2(mock_rados, mock_rbd):
            inst.mock_rados = mock_rados
            inst.mock_rbd = mock_rbd
            inst.mock_rados.ObjectNotFound = MockObjectNotFoundException
            return f(inst, *args, **kwargs)

        return _common_inner_inner2()
    return _common_inner_inner1


class VolumeMetadataBackupTestCase(test.TestCase):

    def setUp(self):
        global RAISED_EXCEPTIONS
        RAISED_EXCEPTIONS = []
        super(VolumeMetadataBackupTestCase, self).setUp()
        self.backup_id = str(uuid.uuid4())
        self.mb = ceph.VolumeMetadataBackup(mock.Mock(), self.backup_id)

    @common_meta_backup_mocks
    def test_name(self):
        self.assertEqual('backup.%s.meta' % (self.backup_id), self.mb.name)

    @common_meta_backup_mocks
    def test_exists(self):
        # True
        self.assertTrue(self.mb.exists)
        self.assertTrue(self.mock_rados.Object.return_value.stat.called)
        self.mock_rados.Object.return_value.reset_mock()

        # False
        self.mock_rados.Object.return_value.stat.side_effect = (
            self.mock_rados.ObjectNotFound)
        self.assertFalse(self.mb.exists)
        self.assertTrue(self.mock_rados.Object.return_value.stat.called)
        self.assertEqual([MockObjectNotFoundException], RAISED_EXCEPTIONS)

    @common_meta_backup_mocks
    def test_set(self):
        obj_data = []
        called = []

        def mock_read(*args):
            called.append('read')
            self.assertTrue(len(obj_data) == 1)
            return obj_data[0]

        def _mock_write(data):
            obj_data.append(data)
            called.append('write')

        self.mb.get = mock.Mock()
        self.mb.get.side_effect = mock_read

        with mock.patch.object(ceph.VolumeMetadataBackup, 'set') as mock_write:
            mock_write.side_effect = _mock_write

            self.mb.set({'foo': 'bar'})
            self.assertEqual({'foo': 'bar'}, self.mb.get())
            self.assertTrue(self.mb.get.called)

            self.mb._exists = mock.Mock()
            self.mb._exists.return_value = True

        # use the unmocked set() method.
        self.assertRaises(exception.VolumeMetadataBackupExists, self.mb.set,
                          {'doo': 'dah'})

        # check the meta obj state has not changed.
        self.assertEqual({'foo': 'bar'}, self.mb.get())

        self.assertEqual(['write', 'read', 'read'], called)

    @common_meta_backup_mocks
    def test_get(self):
        self.mock_rados.Object.return_value.stat.side_effect = (
            self.mock_rados.ObjectNotFound)
        self.mock_rados.Object.return_value.read.return_value = 'meta'
        self.assertIsNone(self.mb.get())
        self.mock_rados.Object.return_value.stat.side_effect = None
        self.assertEqual('meta', self.mb.get())

    @common_meta_backup_mocks
    def remove_if_exists(self):
        with mock.patch.object(self.mock_rados.Object, 'remove') as \
                mock_remove:
            mock_remove.side_effect = self.mock_rados.ObjectNotFound
            self.mb.remove_if_exists()
            self.assertEqual([MockObjectNotFoundException], RAISED_EXCEPTIONS)

            self.mock_rados.Object.remove.side_effect = None
            self.mb.remove_if_exists()
            self.assertEqual([], RAISED_EXCEPTIONS)
